#include <iostream>
#include "chebyshev.hh"
#include "luminosity.hh"
#include "math/integration.hh"
#include <gsl/gsl_math.h>
#include <gsl/gsl_sf.h>

using namespace std;


namespace resumm {
extern double LumRescaled(double x, void *p);
}

namespace cheb {

  int **Tnk;
  unsigned int current_Nmax=0;

  // coefficients T_nk of Chebyshev expansion
  int T(int n, int k) {
    int value=1;
    if(k>n || n<0 || k<0) {
      value = 0;
      //cerr << "Wrong values of (n,k) in T_{n,k}; possible problems are: n<0, k<0, k>n." << endl;
      //exit(0);
    } else if(((n-k)%2)==1) {
      value = 0;
    } else if(n>0) {
      if(k==0) {
	value= -Tnk[n-2][k];
      } else if(k==n && n!=1) {
	value = 2*Tnk[n-1][k-1];
      } else if(k<n) {
	value = 2*Tnk[n-1][k-1]-Tnk[n-2][k];
      }
    }
    return value;
  }

  // this function computes coefficients T[n][k] up to n=N
  int initT(unsigned int N) {
    if(N<=current_Nmax) return 1;
    else {
      current_Nmax = N;
      Tnk = new int*[N+1];
      for(unsigned int n=0; n<=N; n++) {
	Tnk[n]= new int[N+1];
	for(unsigned int k=0; k<=N; k++) {
	  Tnk[n][k]=T(n,k);
	}
      }
      return 1;
    }
    return 0;
  }




  double ctilde(double *c, int k, int N) {
    double value=0;
    for(int n=k; n<=N; n++){
      value += c[n]*cheb::Tnk[n][k];
    }
    if(k==0) value -= 0.5*c[0];
    //if(k==N) value -= 0.5*c[N];  // TEST!!!!
    return value;
  }
  double chat(double *c, int p, int N, double zmax) {
    double value=0;
    for(int k=p; k<=N; k++){
      value += pow(-1.,k-p)*c[k]*gsl_sf_choose(k,p);
    }
    return value*pow(2./zmax,p);
  }
  double bp(double *c, int p, int N, double A) {
    double value=0;
    for(int k=p; k<=N; k++){
      value += c[k]*gsl_sf_choose(k,p);
    }
    return value*pow(-A,p);
  }
  double bp_log1z(double *c, int p, int N, double A) {
    double value=0;
    for(int k=p; k<=N; k++){
      value += c[k]*gsl_sf_choose(k,p)*pow(-1.,k);
    }
    return value*pow(-A,p);
  }
  double cbar(double *c, int p, int N, double M) {
    double value=0;
    for(int k=p; k<=N; k++){
      value += (c[k]*gsl_sf_fact(k))/gsl_sf_fact(k-p);
    }
    return value*pow(-2./M,p);
  }



};

// routine adapted from GSL library
int Chebyshev::cheb_approx(Lum_type type, unsigned int N,
			   const double a, const double b, double *c, void *pars) {
  if(a >= b) {
    cout << "ERROR: in Chebyshev approx, wrong interval [a,b]" << endl;
    abort();
  }
  //
  double bma = 0.5*(b-a);
  double bpa = 0.5*(b+a);
  double fac = 2./(N+1.);
  double *f = new double[N+1];
  //
  for(unsigned int k=0; k<=N; k++) {
    double y = cos(M_PI*(k+0.5)/(N+1.));
    f[k] = Lum_func(type, y*bma + bpa, pars);
  }
  //
  for(unsigned int j=0; j<=N; j++) {
    double sum = 0.;
    for(unsigned int k=0; k<=N; k++) 
      sum += f[k]*cos(M_PI*j*(k+0.5)/(N+1.));
    c[j] = fac * sum;
  }
  return 1;
}






Chebyshev::Chebyshev(resumm::Resummation *r) {
  //res = new resumm::Resummation(r);
  res = r;
  //cheb::initT(N);
  //fLum = Lum;
  //beta_lum=5.;
  //gamma_lum=6.;
  quality_check = false;
  return;
}
Chebyshev::~Chebyshev() {
  return;
}


/*
double Chebyshev::Lum(double z) {
  double res;
  fLum(z, res);
  return res;
}
*/
double Chebyshev::fLum_manipulated(double z, void *p) {
  return resumm::LumRescaled(z,res) * pow(1-z,-gamma_lum) * pow(z,beta_lum);
}
double Chebyshev::fLum_log(double x, void *p) {
  double z = exp(x);
  //return resumm::Lum(z,*res) * z;
  return fLum_manipulated(z,p);
}
double Chebyshev::fLum_plus(double z, void *p) {
  double x = *(double *) p;
  //return (resumm::Lum(x/z,*res)/z - resumm::Lum(x,*res))/(1.-z);
  return (resumm::LumRescaled(x/z,res)/z/resumm::LumRescaled(x,res) - 1.)/(1.-z);
}
double Chebyshev::fLum_plus_log1z(double t, void *p) {
  double x = *(double *) p;
  // z = e^-t
  //return (resumm::Lum(x*exp(t),*res)*exp(t) - resumm::Lum(x,*res))/t;
  return (resumm::LumRescaled(x*exp(t),res)*exp(t)/resumm::LumRescaled(x,res) - 1.)/t;
}

double Chebyshev::Lum_func(Lum_type type, double z, void *p) {
  switch(type) {
    //case Lum_manip:
    //return fLum_manipulated(z,p);
  case Lum_log:
    return fLum_log(z,p);
  case Lum_plus:
    return fLum_plus(z,p);
  case Lum_plus_log1z:
    return fLum_plus_log1z(z,p);
  }
  abort();
  return 0;
}


/*
// reconstruction of original functions (used for Quality Check)
double Chebyshev::rec_Lum_plus(double *c, int N, double z) {
  double value=0;
  for(int k=0; k<=N; k++){
    value += c[k]*pow(1.-z,k);
  }
  return value;
}
double Chebyshev::rec_Lum_plus_log1z(double *c, int N, double z) {
  double value=0;
  for(int k=0; k<=N; k++){
    value += c[k]*pow(-log(z),k);
  }
  return value;
}
double Chebyshev::rec_Lum(double *c, int N, double z) {
  double value = 0;
  for(int i=0; i<=N; i++) {
    for(int j=0; j<=gamma_lum; j++) {
      value += c[i] * pow(z,j-beta_lum)* pow(-1.,j) * gsl_sf_choose(gamma_lum,j) *pow(-log(z),i)/gsl_sf_fact(i);
    }
  }
  return value;
}
double Chebyshev::rec_Lum_func(Lum_type type, double z, double *c, int N) {
  switch(type) {
    //case Lum_manip:
    //return fLum_manipulated(z,p);
  case Lum_log:
    return rec_Lum(c,N,z);
  case Lum_plus:
    return rec_Lum_plus(c,N,z);
  case Lum_plus_log1z:
    return rec_Lum_plus_log1z(c,N,z);
  }
  abort();
  return 0;
}
*/
double recLum_func(double u, double *c, int N, double A, double B) {
  double res = 0.;
  for(int i=0; i<=N; i++) {
    res += c[i] * pow(A*u+B, i);
  }
  return res;
}
struct sQC {
  Chebyshev *cheb;
  Lum_type type;
  void *extra_pars;
  int N;
  double *c;
  double A, B;
};
double integrand_QC(double u, void *p) {
  sQC pars = *(sQC *) p;
  if(pars.type == Lum_plus) u = exp(-u);
  return fabs( pars.cheb->Lum_func(pars.type, u, pars.extra_pars)
	       //- pars.cheb->rec_Lum_func(pars.type, z, pars.c, pars.N);
	       - recLum_func(u, pars.c, pars.N, pars.A, pars.B)
	       );
}





/*
int Chebyshev::ApproxLuminosity(double *coeff, unsigned int N, double beta, double gamma) {
  using namespace cheb;
  initT(N);
  beta_lum = beta;
  gamma_lum = gamma;
  //
  double *coeff_lum_tmp, *coeff_tilde_lum;
  coeff_lum_tmp = new double[N+1];
  coeff_tilde_lum = new double[N+1];
  cheb_approx(Lum_manip, N, 0., 1., coeff_lum_tmp, NULL);
  //
  for(unsigned int k=0; k<=N; k++) {
    coeff_tilde_lum[k] = ctilde(coeff_lum_tmp,k,N);
  }
  for(unsigned int k=0; k<=N; k++) {
    coeff[k] = chat(coeff_tilde_lum,k,N,1.);
  }
  return 1;
}
*/
int Chebyshev::ApproxLuminosity(double *coeff, unsigned int N, double beta, int gamma, double z_min) {
  using namespace cheb;
  initT(N);
  beta_lum = beta;
  gamma_lum = gamma;
  //
  double M=-log(z_min);
  double *coeff_tmp   = new double[N+1];
  double *coeff_tilde = new double[N+1];
  cheb_approx(Lum_log, N, -M, 0., coeff_tmp, NULL);
  //
  for(unsigned int k=0; k<=N; k++) {
    coeff_tilde[k] = ctilde(coeff_tmp,k,N);
  }
  for(unsigned int k=0; k<=N; k++) {
    coeff[k] = cbar(coeff_tilde,k,N,M);
  }
  if(quality_check) {
    double prec = 1.e-1;
    sQC pars = { this, Lum_log, NULL, N, coeff_tilde, 2./M, 1. };
    double qc = integration::gauss(integrand_QC, -M, 0., prec, NULL, &pars);
    cout << "Quality Check: this number should be as small as possible: " << qc << endl;
  }
  return 1;
}

int Chebyshev::ApproxLumPlus(double *coeff, double x, unsigned int N) {
  using namespace cheb;
  initT(N);
  //
  double *coeff_tmp   = new double[N+1];
  double *coeff_tilde = new double[N+1];
  double A = 2./(1.-x);
  cheb_approx(Lum_plus, N, x, 1., coeff_tmp, &x);
  //
  for(unsigned int i=0; i<=N; i++){
    coeff_tilde[i] = ctilde(coeff_tmp,i,N);
  }
  for(unsigned int i=0; i<=N; i++){
    coeff[i] = bp(coeff_tilde,i,N,A);
  }
  if(quality_check) {
    double prec = 1.e-1;
    sQC pars = { this, Lum_plus, &x, N, coeff_tilde, A, 1.-A };
    //double qc = integration::gauss(integrand_QC, x, 1., prec, NULL, &pars);
    double qc = integration::gauss(integrand_QC, 0., -log(x), prec, NULL, &pars);
    cout << "Quality Check: this number should be as small as possible: " << qc << endl;
  }
  return 1;
}

int Chebyshev::ApproxLumPlus_log1z(double *coeff, double x, unsigned int N) {
  using namespace cheb;
  initT(N);
  //
  double *coeff_tmp   = new double[N+1];
  double *coeff_tilde = new double[N+1];
  //double a=6, tmax=a*log(10.);
  double tmax=-log(x);
  double A = 2./tmax;
  cheb_approx(Lum_plus_log1z, N, 0., tmax, coeff_tmp, &x);
  //
  for(unsigned int i=0; i<=N; i++){
    coeff_tilde[i] = ctilde(coeff_tmp,i,N);
  }
  for(unsigned int i=0; i<=N; i++){
    coeff[i] = bp_log1z(coeff_tilde,i,N,A);
  }
  if(quality_check) {
    double prec = 1.e-1;
    sQC pars = { this, Lum_plus_log1z, &x, N, coeff_tilde, A, -1. };
    double qc = integration::gauss(integrand_QC, 0., tmax, prec, NULL, &pars);
    cout << "Quality Check: this number should be as small as possible: " << qc << endl;
  }
  return 1;
}

void Chebyshev::EnableQualityCheck(bool enable) {
  quality_check = enable;
}
