#include "luminosity.hh"
#include "QCD_EW.hh"
#include "math/integration.hh"
#include "PDFinterface.hh"


using namespace std;
using namespace PDFinterface;



namespace lumi {

  // Higgs part
  double lumi_gg(PDFs *pdfs, double x1, double x2, double mu) {
    return pdfs->xfx(x1,mu,0)/x1 * pdfs->xfx(x2,mu,0)/x2;
  }
  
	
  // DIS part
  const double a_u_Z = -1., a_d_Z = 1.;
  const double v_u_Z = 1.-8./3.*EW::s2W, v_d_Z = -1.+4./3.*EW::s2W;
  const double A_e = -1./2;
  const double V_e = -1./2 + 2.*EW::s2W;
  const double A_u = 1./2;
  const double V_u = 1./2 - 4./3*EW::s2W;
  const double A_d = -1./2;
  const double V_d = -1./2 + 2./3*EW::s2W;

  double Zcoupling = 1;  // set to 0 to exclude the Z
  void SetZcoupling(double c) {
    Zcoupling = c;
  }
  double P_z(double mu) {
    double musq = mu*mu;
    double mZsq = MASS::mass_Z*MASS::mass_Z;
    return Zcoupling/(4.*EW::s2W*(1.-EW::s2W))*(musq/(musq+mZsq));
  }

  bool aF2 = 1;
  bool aF3 = 1;
  void SetDISObservable(DISobservable obs) {
    if(obs == F2)    { aF2 = 1; aF3 = 0; }
    if(obs == F3)    { aF2 = 0; aF3 = 1; }
    if(obs == sigma) { aF2 = 1; aF3 = 1; }
  }

  // neutral current
  double lumi_DISn(PDFs *pdfs, double x, double mu, int nf, collider c, double y, int lepton_charge, double TMtau, bool integ) {
    double xfx[13];
    if     (c==proton)   pdfs->xfx(x, mu, xfx);
    else if(c==deuteron) pdfs->xfx(x, mu, xfx, Deuteron);
    else if(c==iron)     pdfs->xfx(x, mu, xfx, Iron);
    else abort();
    double pZ = P_z(mu);
    double pref_u_2 = EW::e_u*EW::e_u - 2*V_e*pZ*EW::e_u*V_u + (V_e*V_e+A_e*A_e)*pZ*pZ*(V_u*V_u+A_u*A_u);
    double pref_d_2 = EW::e_d*EW::e_d - 2*V_e*pZ*EW::e_d*V_d + (V_e*V_e+A_e*A_e)*pZ*pZ*(V_d*V_d+A_d*A_d);
    double pref_u_3 = -2*A_e*pZ*EW::e_u*A_u + 4*V_e*A_e*pZ*pZ*V_u*A_u;
    double pref_d_3 = -2*A_e*pZ*EW::e_d*A_d + 4*V_e*A_e*pZ*pZ*V_d*A_d;
    double Yplus  = 1. + pow(1.-y,2);
    double Yminus = 1. - pow(1.-y,2);
    Yplus  -= y*y *(TMtau*TMtau-1.)/2.;
    if(aF2==0) { Yplus = 0; Yminus = 1; }
    if(aF3==0) { Yplus = 1; Yminus = 0; }
    Yminus *= TMtau;
    if(integ) {
      Yplus  *= 3;
      Yminus *= 2;
    }
    //cout << y << " " << Yplus << " " << Yminus << endl;
    double pref_u  = Yplus*pref_u_2 - lepton_charge* Yminus*pref_u_3;
    double pref_ub = Yplus*pref_u_2 + lepton_charge* Yminus*pref_u_3;
    double pref_d  = Yplus*pref_d_2 - lepton_charge* Yminus*pref_d_3;
    double pref_db = Yplus*pref_d_2 + lepton_charge* Yminus*pref_d_3;
    double S=0;
    for(int i=1; i<=nf; i++) {
      //if(i%2==0) S += pref_u_2 * (xfx[6+i]+xfx[6-i]);
      //if(i%2==1) S += pref_d_2 * (xfx[6+i]+xfx[6-i]);
      if(i%2==0) S += pref_u*xfx[6+i] + pref_ub*xfx[6-i];
      if(i%2==1) S += pref_d*xfx[6+i] + pref_db*xfx[6-i];
    }
    //if(integ) cout << setw(12) << x << "  " << setw(12) << S/x << endl;
    return S/x;
  }
  struct DISnintpars {PDFs *pdfs; double mu; int nf; collider c; double y; int lepton_charge; double tau; double mcoQ; double x; };
  double lumi_DISn_integrand(double t, void *par) {
    DISnintpars p = *(DISnintpars *) par;
    double x = p.x;
    double w = exp(t*log(x));
    return -log(x) * lumi_DISn(p.pdfs, w, p.mu, p.nf, p.c, p.y, p.lepton_charge, p.tau, true) / w;
  }
  double lumi_DISn(PDFs *pdfs, double x, double mu, int nf, collider c, double y, int lepton_charge, double TMtau) {
    double res = lumi_DISn(pdfs, x, mu, nf, c, y, lepton_charge, TMtau, false);
    if(TMtau>1) {
      DISnintpars pars = { pdfs, mu, nf, c, y, lepton_charge, TMtau, 0, x };
      double precision = 2e-3;
      res += (TMtau-1)/TMtau*integration::gauss(lumi_DISn_integrand, 0, 1, precision, NULL, &pars);
    }
    return res;
  }




  // charged current
  double lumi_DISc(PDFs *pdfs, double x, double mu, int nf, collider c, double y, bool incoming_nu, int Wcharge, double TMtau, bool integ) {
    double xfx[13];
    if     (c==proton)   pdfs->xfx(x, mu, xfx);
    else if(c==deuteron) pdfs->xfx(x, mu, xfx, Deuteron);
    else if(c==iron)     pdfs->xfx(x, mu, xfx, Iron);
    else abort();
    double Yplus  = 1. + pow(1.-y,2);
    double Yminus = 1. - pow(1.-y,2);
    Yplus  -= y*y *(TMtau*TMtau-1.)/2.;
    if(aF2==0) { Yplus = 0; Yminus = 1; }
    if(aF3==0) { Yplus = 1; Yminus = 0; }
    Yminus *= TMtau;
    if(integ) {
      Yplus  *= 3;
      Yminus *= 2;
    }
    double pref_u = Yplus - (incoming_nu?1:-1)* Yminus;
    double pref_d = Yplus + (incoming_nu?1:-1)* Yminus;
    double S=0;
    double V2CKMsum[6];
    //const string sVCKM[3][3] = { {"ud", "us", "ub"}, {"cd", "cs", "cb"}, {"td", "ts", "tb"} };
    for(int i=0; i<6; i++) {  // note that i=0 is d, i=1 is u, ... (shift by one)
      V2CKMsum[i] = 0;
      if(i<nf) {
	for(int j=0; j<3; j++) {
	  if(i%2==0 && 2*(j+1)<=nf) {  // d-type quark
	    V2CKMsum[i] += EW::V2CKM[j][i/2];
	    //cout << sVCKM[j][i/2] << "   ";
	  }
	  if(i%2==1 && 2*(j+1)-1<=nf) {  // u-type quark
	    V2CKMsum[i] += EW::V2CKM[i/2][j];
	    //cout << sVCKM[i/2][j] << "   ";
	  }
	}
      }
      //cout << V2CKMsum[i] << endl;
    }
    //
    for(int i=1; i<=nf; i++) {
      if(i%2==0) S += pref_u * V2CKMsum[i-1] * xfx[6-Wcharge*i];
      if(i%2==1) S += pref_d * V2CKMsum[i-1] * xfx[6+Wcharge*i];
    }
    return S/x*2;
  }
  struct DIScintpars {PDFs *pdfs; double mu; int nf; collider c; double y; bool incoming_nu; int Wcharge; double tau; double mcoQ; double x; };
  double lumi_DISc_integrand(double t, void *par) {
    DIScintpars p = *(DIScintpars *) par;
    double x = p.x;
    double w = exp(t*log(x));
    return -log(x) * lumi_DISc(p.pdfs, w, p.mu, p.nf, p.c, p.y, p.incoming_nu, p.Wcharge, p.tau, true) / w;
  }
  double lumi_DISc(PDFs *pdfs, double x, double mu, int nf, collider c, double y, bool incoming_nu, int Wcharge, double TMtau) {
    double res = lumi_DISc(pdfs, x, mu, nf, c, y, incoming_nu, Wcharge, TMtau, false);
    if(TMtau>1) {
      DIScintpars pars = { pdfs, mu, nf, c, y, incoming_nu, Wcharge, TMtau, 0, x };
      double precision = 2e-3;
      res += (TMtau-1)/TMtau*integration::gauss(lumi_DISc_integrand, 0, 1, precision, NULL, &pars);
    }
    return res;
  }




  double lumi_DIScharmn(PDFs *pdfs, double x, double mu, int nf, collider c, double y, int lepton_charge, double TMtau, double mcoQ, bool integ) {
    if(x>1) return 0;
    double xfx[13];
    if     (c==proton)   pdfs->xfx(x, mu, xfx);
    else if(c==deuteron) pdfs->xfx(x, mu, xfx, Deuteron);
    else if(c==iron)     pdfs->xfx(x, mu, xfx, Iron);
    else abort();
    double pZ = P_z(mu);
    double pref_u_2 = EW::e_u*EW::e_u - 2*V_e*pZ*EW::e_u*V_u + (V_e*V_e+A_e*A_e)*pZ*pZ*(V_u*V_u+A_u*A_u);
    double pref_u_3 = -2*A_e*pZ*EW::e_u*A_u + 4*V_e*A_e*pZ*pZ*V_u*A_u;
    double Yplus  = 1. + pow(1.-y,2);
    double Yminus = 1. - pow(1.-y,2);
    Yplus  -= y*y *(TMtau*TMtau-1.)/2. + y*y;
    Yplus  *= 1.+mcoQ*mcoQ;
    Yplus  += y*y;
    if(aF2==0) { Yplus = 0; Yminus = 1; }
    if(aF3==0) { Yplus = 1; Yminus = 0; }
    Yminus *= TMtau;
    if(integ) {
      Yplus  *= 3;
      Yminus *= 2;
    }
    double pref_u  = Yplus*pref_u_2 - lepton_charge* Yminus*pref_u_3;
    double pref_ub = Yplus*pref_u_2 + lepton_charge* Yminus*pref_u_3;
    double S = pref_u*xfx[6+4] + pref_ub*xfx[6-4];
    //double pref_u = EW::e_u*EW::e_u - V_e*pZ*(2.*EW::e_u*V_u) +(V_e*V_e+A_e*A_e)*pZ*pZ*(V_u*V_u+A_u*A_u);
    //double S = pref_u * (xfx[6+4]+xfx[6-4]);
    return S/x;
  }
  double lumi_DIScharmn_integrand(double t, void *par) {
    DISnintpars p = *(DISnintpars *) par;
    double x = p.x;
    double w = exp(t*log(x));
    return -log(x) * lumi_DIScharmn(p.pdfs, w, p.mu, p.nf, p.c, p.y, p.lepton_charge, p.tau, p.mcoQ, true) / w;
  }
  double lumi_DIScharmn(PDFs *pdfs, double x, double mu, int nf, collider c, double y, int lepton_charge, double TMtau, double mcoQ) {
    double res = lumi_DIScharmn(pdfs, x, mu, nf, c, y, lepton_charge, TMtau, mcoQ, false);
    if(TMtau>1) {
      DISnintpars pars = { pdfs, mu, nf, c, y, lepton_charge, TMtau, mcoQ, x };
      double precision = 2e-3;
      res += (TMtau-1)/TMtau*integration::gauss(lumi_DIScharmn_integrand, 0, 1, precision, NULL, &pars);
    }
    return res;
  }



  double lumi_DIScharmc(PDFs *pdfs, double x, double mu, int nf, collider c, double y, bool incoming_nu, int Wcharge, double TMtau, double mcoQ, bool integ) {
    if(x>1) return 0;
    double xfx[13];
    if     (c==proton)   pdfs->xfx(x, mu, xfx);
    else if(c==deuteron) pdfs->xfx(x, mu, xfx, Deuteron);
    else if(c==iron)     pdfs->xfx(x, mu, xfx, Iron);
    else abort();
    double V_CKM_c[6];
    V_CKM_c[1] = EW::vcd*EW::vcd;
    V_CKM_c[3] = EW::vcs*EW::vcs;
    V_CKM_c[5] = EW::vcb*EW::vcb;
    double Yplus  = 1. + pow(1.-y,2);
    double Yminus = 1. - pow(1.-y,2);
    Yplus  -= y*y *(TMtau*TMtau-1.)/2. + y*y;
    Yplus  *= 1.+mcoQ*mcoQ;
    Yplus  += y*y;
    if(aF2==0) { Yplus = 0; Yminus = 1; }
    if(aF3==0) { Yplus = 1; Yminus = 0; }
    Yminus *= TMtau;
    if(integ) {
      Yplus  *= 3;
      Yminus *= 2;
    }
    //double pref_u = Yplus - (incoming_nu?1:-1)* Yminus;
    double pref_d = Yplus + (incoming_nu?1:-1)* Yminus;
    double S=0;
    for(int i=1; i<=nf; i++) {
      //if(i%2==1) S += V_CKM_c[(i-1)/2]*pref_d * xfx[6+Wcharge*i];
      if(i%2==1) S += V_CKM_c[i]*pref_d * xfx[6+Wcharge*i];
    }
    return S/x*2;
  }
  double lumi_DIScharmc_integrand(double t, void *par) {
    DIScintpars p = *(DIScintpars *) par;
    double x = p.x;
    double w = exp(t*log(x));
    return -log(x) * lumi_DIScharmc(p.pdfs, w, p.mu, p.nf, p.c, p.y, p.incoming_nu, p.Wcharge, p.tau, p.mcoQ, true) / w;
  }
  double lumi_DIScharmc(PDFs *pdfs, double x, double mu, int nf, collider c, double y, bool incoming_nu, int Wcharge, double TMtau, double mcoQ) {
    double res = lumi_DIScharmc(pdfs, x, mu, nf, c, y, incoming_nu, Wcharge, TMtau, mcoQ, false);
    if(TMtau>1) {
      DIScintpars pars = { pdfs, mu, nf, c, y, incoming_nu, Wcharge, TMtau, mcoQ, x };
      double precision = 2e-3;
      res += (TMtau-1)/TMtau*integration::gauss(lumi_DIScharmc_integrand, 0, 1, precision, NULL, &pars);
    }
    return res;
  }



  double lumi_DIS(PDFs *pdfs, double x, double mu, int nf, process p, collider c, double y, double TMtau, double mcoQ) {
    double res;
    if     (p == DISneutral     ) res = lumi_DISn(pdfs, x,mu,nf,c,y, 1,TMtau) + lumi_DISn(pdfs, x,mu,nf,c,y,-1,TMtau);
    else if(p == DISneutralPlus ) res = lumi_DISn(pdfs, x,mu,nf,c,y, 1,TMtau);
    else if(p == DISneutralMinus) res = lumi_DISn(pdfs, x,mu,nf,c,y,-1,TMtau);
    else if(p == DISchargedWpnu) res = lumi_DISc(pdfs, x,mu,nf,c,y,true , 1,TMtau);
    else if(p == DISchargedWmnu) res = lumi_DISc(pdfs, x,mu,nf,c,y,true ,-1,TMtau);
    else if(p == DISchargedWpe ) res = lumi_DISc(pdfs, x,mu,nf,c,y,false, 1,TMtau);
    else if(p == DISchargedWme ) res = lumi_DISc(pdfs, x,mu,nf,c,y,false,-1,TMtau);
    else if(p == DISneutralCharm     ) res = lumi_DIScharmn(pdfs, x,mu,nf,c,y, 1,TMtau,mcoQ) + lumi_DIScharmn(pdfs, x,mu,nf,c,y,-1,TMtau,mcoQ);
    else if(p == DISneutralCharmPlus ) res = lumi_DIScharmn(pdfs, x,mu,nf,c,y, 1,TMtau,mcoQ);
    else if(p == DISneutralCharmMinus) res = lumi_DIScharmn(pdfs, x,mu,nf,c,y,-1,TMtau,mcoQ);
    else if(p == DISchargedCharmWpnu) res = lumi_DIScharmc(pdfs, x,mu,nf,c,y,true , 1,TMtau,mcoQ);
    else if(p == DISchargedCharmWmnu) res = lumi_DIScharmc(pdfs, x,mu,nf,c,y,true ,-1,TMtau,mcoQ);
    else if(p == DISchargedCharmWpe ) res = lumi_DIScharmc(pdfs, x,mu,nf,c,y,false, 1,TMtau,mcoQ);
    else if(p == DISchargedCharmWme ) res = lumi_DIScharmc(pdfs, x,mu,nf,c,y,false,-1,TMtau,mcoQ);
    else abort();
    return res;
  }




  // Drell-Yan part
  //
  const double V_ud = EW::vud;//0.975;
  const double V_us = EW::vus;//0.222;
  const double V_cd = EW::vcd;//0.222;
  const double V_cs = EW::vcs;//0.974;
  const double V_ud_sq = V_ud*V_ud;     double V_us_sq = V_us*V_us;
  const double V_cd_sq = V_cd*V_cd;     double V_cs_sq = V_cs*V_cs;
  const double V_ub_sq = 1. - V_ud_sq - V_us_sq;
  const double V_cb_sq = 1. - V_cd_sq - V_cs_sq;
  const double V_td_sq = 1. - V_ud_sq - V_cd_sq;
  const double V_ts_sq = 1. - V_us_sq - V_cs_sq;
  const double V_tb_sq = 1. - V_ub_sq - V_cb_sq;
  //
  double N_Zgamma(double Q, double alpha){
    double Qsq = Q*Q;  double mZsq = MASS::mass_Z*MASS::mass_Z;
    return Zcoupling*((1.-4.*EW::s2W)/8./EW::s2W/EW::c2W * Qsq*(Qsq-mZsq)/((Qsq-mZsq)*(Qsq-mZsq) + EW::Gamma_Z*EW::Gamma_Z * mZsq));
  }
  double N_Z(double Q, double alpha){
    double Qsq = Q*Q;  double mZsq = MASS::mass_Z*MASS::mass_Z;
    return Zcoupling*Zcoupling*(3./16./EW::s2W/EW::c2W/alpha * EW::Gamma_Z * EW::Br_l_Z/MASS::mass_Z
				* Qsq*Qsq/((Qsq-mZsq)*(Qsq-mZsq) + EW::Gamma_Z*EW::Gamma_Z * mZsq));
  }
  double VsqAsq_u(double M, double alpha){
    return EW::e_u*EW::e_u + N_Zgamma(M,alpha)*EW::e_u*v_u_Z + N_Z(M,alpha)*(v_u_Z*v_u_Z + a_u_Z*a_u_Z);
  }
  double VsqAsq_d(double M, double alpha){
    return EW::e_d*EW::e_d + N_Zgamma(M,alpha)*EW::e_d*v_d_Z + N_Z(M,alpha)*(v_d_Z*v_d_Z + a_d_Z*a_d_Z) ;
  }
  double lumi_DYn(PDFs *pdfs, double x1, double x2, double mu, double M, collider c) { // neutral (Z+gamma)
    //double alpha = EW::alpha_QED(M);
    double alpha = EW::alpha_em_Z;
    double pref_u = VsqAsq_u(M,alpha);
    double pref_d = VsqAsq_d(M,alpha);
    double xfx1[13], xfx2[13];
    pdfs->xfx(x1, mu, xfx1);
    if     (c==pp)    pdfs->xfx(x2, mu, xfx2);
    else if(c==ppbar) pdfs->xfx(x2, mu, xfx2, Proton, true);
    else if(c==pd)    pdfs->xfx(x2, mu, xfx2, Deuteron);
    double res = 0;
    for(int i=1; i<=6; i++) {
      if(i%2==0) res += pref_u * (xfx1[6+i]*xfx2[6-i] + xfx1[6-i]*xfx2[6+i]);
      if(i%2==1) res += pref_d * (xfx1[6+i]*xfx2[6-i] + xfx1[6-i]*xfx2[6+i]);
    }
    return res/x1/x2;
  }
  double lumi_DYp(PDFs *pdfs, double x1, double x2, double mu, double M, collider c) { // positive (W+)
    double xfx1[13], xfx2[13];
    pdfs->xfx(x1, mu, xfx1);
    if     (c==pp)    pdfs->xfx(x2, mu, xfx2);
    else if(c==ppbar) pdfs->xfx(x2, mu, xfx2, Proton, true);
    else if(c==pd)    pdfs->xfx(x2, mu, xfx2, Deuteron);
    double res = 0;
    res = 0
      + xfx1[6+2] * ( V_ud_sq * xfx2[6-1] + V_us_sq * xfx2[6-3] + V_ub_sq * xfx2[6-5] )
      + xfx1[6+4] * ( V_cd_sq * xfx2[6-1] + V_cs_sq * xfx2[6-3] + V_cb_sq * xfx2[6-5] )
      + xfx1[6+6] * ( V_td_sq * xfx2[6-1] + V_ts_sq * xfx2[6-3] + V_tb_sq * xfx2[6-5] )
      + xfx2[6+2] * ( V_ud_sq * xfx1[6-1] + V_us_sq * xfx1[6-3] + V_ub_sq * xfx1[6-5] )
      + xfx2[6+4] * ( V_cd_sq * xfx1[6-1] + V_cs_sq * xfx1[6-3] + V_cb_sq * xfx1[6-5] )
      + xfx2[6+6] * ( V_td_sq * xfx1[6-1] + V_ts_sq * xfx1[6-3] + V_tb_sq * xfx1[6-5] );
    return res/x1/x2;
  }
  double lumi_DYm(PDFs *pdfs, double x1, double x2, double mu, double M, collider c) { // negative (W-)
    double xfx1[13], xfx2[13];
    pdfs->xfx(x1, mu, xfx1);
    if     (c==pp)    pdfs->xfx(x2, mu, xfx2);
    else if(c==ppbar) pdfs->xfx(x2, mu, xfx2, Proton, true);
    else if(c==pd)    pdfs->xfx(x2, mu, xfx2, Deuteron);
    double res = 0;
    res = 0
      + xfx1[6-2] * ( V_ud_sq * xfx2[6+1] + V_us_sq * xfx2[6+3] + V_ub_sq * xfx2[6+5] )
      + xfx1[6-4] * ( V_cd_sq * xfx2[6+1] + V_cs_sq * xfx2[6+3] + V_cb_sq * xfx2[6+5] )
      + xfx1[6-6] * ( V_td_sq * xfx2[6+1] + V_ts_sq * xfx2[6+3] + V_tb_sq * xfx2[6+5] )
      + xfx2[6-2] * ( V_ud_sq * xfx1[6+1] + V_us_sq * xfx1[6+3] + V_ub_sq * xfx1[6+5] )
      + xfx2[6-4] * ( V_cd_sq * xfx1[6+1] + V_cs_sq * xfx1[6+3] + V_cb_sq * xfx1[6+5] )
      + xfx2[6-6] * ( V_td_sq * xfx1[6+1] + V_ts_sq * xfx1[6+3] + V_tb_sq * xfx1[6+5] );
    return res/x1/x2;
  }
  //
  double lumi_DY(PDFs *pdfs, double x1, double x2, double mu, double M, process p, collider c) {
    double res;
    if(p == DYneutral) res = lumi_DYn(pdfs, x1,x2,mu,M,c);
    else if(p == DYpositive) res = lumi_DYp(pdfs, x1,x2,mu,M,c);
    else if(p == DYnegative) res = lumi_DYm(pdfs, x1,x2,mu,M,c);
    else abort();
    return res;
  }


  //double NFZgamma(double mu) {
  //}

  double NFV(double mu, process p) {
    double res = 0;
    //if     (p == DYneutral ) res = NFZgamma(mu);
    //else if(p == DYpositive) res = NFWp(mu);
    //else if(p == DYnegative) res = NFWm(mu);
    //else abort();
    return res;
  }


};


