#include <cmath>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>
#include <cuba.h>
#include "resummation.hh"
#include "chebyshev.hh"
#include "lagrange.hh"
#include "QCD_EW.hh"
#include "math/special_functions.hh"
#include "math/integration.hh"
#include "math/struct_definitions.hh"
#include "math/mellin_Dk.hh"
#include "higgs/higgs_consts.hh"


using namespace std;
using namespace PDFinterface;




namespace resumm {

  string sPrescr(prescription p) {
    string s;
    if(p == BP00) s = "BP00";
    if(p == BP0)  s = "BP0";
    if(p == BP1)  s = "BP1";
    if(p == BP2)  s = "BP2";
    if(p == MP00) s = "MP00";
    if(p == MP0)  s = "MP0";
    if(p == MP2)  s = "MP2";
    return s;
  }
  string sAP(APtype ap) {
    string s;
    if(ap == AP0)  s = "AP0";
    if(ap == AP1)  s = "AP1";
    if(ap == AP2)  s = "AP2";
    if(ap == AP )  s = "AP";
    return s;
  }


  ////////////////////////////////
  //        PUBLIC  PART        //
  ////////////////////////////////

  Resummation::Resummation(process p, collider c, string PDFset, int PDFmember) {
    proc = p;
    coll = c;
    pdfs = new PDFs(PDFset, PDFmember);
    InitMasses        (pdfs->getQuarkMass     (4), pdfs->getQuarkMass     (5), pdfs->getQuarkMass     (6));
    InitMassThresholds(pdfs->getQuarkThreshold(4), pdfs->getQuarkThreshold(5), pdfs->getQuarkThreshold(6));
    Hmtop = pdfs->getQuarkMass(6);
    Hmbot = -1;
    Hmcha = -1;
    mtEFT = true;
    TargetMass = 0;
    DISy = 0;
    asrescale = 1;
    //
    // default setup: psi-soft_2 with g0 NOT exponentiated
    prescr = MP2;
    APimpr = AP2;
    g0exp = false;
    expDconst = false;
    expBP = false;
    //
    // other init...
    SetBPpars();
    SetMPpars();
    interpolationMethod = 0;
    LogarithmicCounting = primed;
    DampingOrder = 0;
    doNsubtraction = false;
    //
    last_muFrat = __builtin_nan("");
    last_muRrat = __builtin_nan("");
    last_M      = __builtin_nan("");
    last_Y      = __builtin_nan("");
    last_sqrts  = __builtin_nan("");
    last_nf     = __builtin_nan("");
    last_mtop   = __builtin_nan("");
    last_EFT    = __builtin_nan("");
    last_Ncheb  = __builtin_nan("");
    firstinit = true;
    g0constComputation = false;
    forceredo = false;
    //
    return;
  }

  Resummation::~Resummation() {
    return;
  }

  void Resummation::reinitPDF(string PDFset, int PDFsubset) {
    delete pdfs;
    pdfs = new PDFs(PDFset, PDFsubset);
    InitMasses        (pdfs->getQuarkMass     (4), pdfs->getQuarkMass     (5), pdfs->getQuarkMass     (6));
    InitMassThresholds(pdfs->getQuarkThreshold(4), pdfs->getQuarkThreshold(5), pdfs->getQuarkThreshold(6));
    Hmtop = pdfs->getQuarkMass(6);
  }

  void Resummation::SetMtop(double mt) {
    Hmtop = mt;
  }
  void Resummation::SetMbottom(double mb) {
    Hmbot = mb;
  }
  void Resummation::SetMcharm(double mc) {
    Hmcha = mc;
  }
  void Resummation::Setmuf(double mu) {
    muF = mu;
  }
  void Resummation::UseLargeMtopEFT(bool use) {
    mtEFT = use;
  }
  void Resummation::SetTargetMass(double target_mass) {  // has an effect only for DIS; in GeV
    TargetMass = target_mass;
  }

  void Resummation::SetPrescriptionByName(string pr, bool exponentiate_g0) {
    g0exp = exponentiate_g0;
    expBP = false;
    expDconst = false;
    if(pr.substr(0,6) == "A-soft") {
      expBP = true;
      expDconst = true;
      prescr = BP2;
      SetBPpars();
      string prN = pr.substr(6,pr.length()-6);
      if(prN == "_0" || prN == "0" || prN == "") APimpr = AP0;
      if(prN == "_1" || prN == "1") APimpr = AP1;
      if(prN == "_2" || prN == "2") APimpr = AP2;
    } else if(pr.substr(0,8) == "psi-soft") {
      prescr = MP2;
      SetMPpars();
      string prN = pr.substr(8,pr.length()-8);
      if(prN == "_0" || prN == "0" || prN == "") APimpr = AP0;
      if(prN == "_1" || prN == "1") APimpr = AP1;
      if(prN == "_2" || prN == "2") APimpr = AP2;
    } else if(pr.substr(0,8) == "Psi-soft") {
      expDconst = true;
      prescr = MP2;
      SetMPpars();
      string prN = pr.substr(8,pr.length()-8);
      if(prN == "_0" || prN == "0" || prN == "") APimpr = AP0;
      if(prN == "_1" || prN == "1") APimpr = AP1;
      if(prN == "_2" || prN == "2") APimpr = AP2;
    } else if(pr.substr(0,6) == "N-soft" && pr.length() == 6) {
      prescr = MP00;
      SetMPpars();
      APimpr = AP0;
    } else if(pr.substr(0,6) == "z-soft" && pr.length() == 6) {
      prescr = BP1;
      SetBPpars();
      APimpr = AP0;
    } else {
      cerr << "The selected prescription does not exist!" << endl;
      abort();
    }
  }
  void Resummation::SetPrescription(prescription p, bool exponentiate_g0, bool resumDistrConst, bool expBorel) {
    prescr = p;
    if(p < 0) SetMPpars();
    else SetBPpars();
    //shift = PSfactor;
    g0exp = exponentiate_g0;
    expDconst = resumDistrConst;
    if(expBorel && prescr>0) expBP=true;
  }
  void Resummation::SetAPimprovement(APtype impr) {
    APimpr = impr;
  }
  void Resummation::SetInterpolationMethod(string name) {
    if(name.substr(0,9) == "Chebyshev") interpolationMethod = 0; else
    if(name.substr(0,8) == "Lagrange" ) interpolationMethod = 1;
    else {
      cerr << "The selected interpolation method does not exist!" << endl;
      abort();
    }
  }
  void Resummation::SetBPpars(unsigned int Ncheb, double cutoff) {
    if(expBP) SetMPpars(Ncheb);
    Nchebyshev = Ncheb;
    BPcutoff = cutoff;
  }
  void Resummation::SetMPpars(unsigned int Ncheb, double beta, int gamma, double z_min) {
    Nchebyshev = Ncheb;
    MPbeta = beta;
    MPgamma = gamma;
    MPzmin = z_min;
  }
  void Resummation::SetMPpars(unsigned int Ncheb) {
    SetMPpars(Ncheb, 1., 0, 1.e-3);
  }

  void Resummation::SetLogCounting(LogCountType lct) {
    LogarithmicCounting = lct;
  }
  
  void Resummation::SetDamping(unsigned int order) {
    DampingOrder = order;
  }
  void Resummation::SetConstSubtraction(dcomplex N) {
    doNsubtraction = true;
    Nsubtraction = N;
  }
  void Resummation::UnsetConstSubtraction() {
    doNsubtraction = false;
  }

  double Resummation::alphas(double mu) {
    return pdfs->alphasQ(mu);
  }

  PDFinterface::PDFs *Resummation::GetPDFs() {
    return pdfs;
  }


  double Resummation::DISxsLL(double x, double Q2, double y, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    DISy = y;
    return res_xs(sqrt(Q2), sqrt(Q2/x), muF_m_ratio, muR_m_ratio, LL, match_order, asinput);
  }
  double Resummation::DISxsNLL(double x, double Q2, double y, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    DISy = y;
    return res_xs(sqrt(Q2), sqrt(Q2/x), muF_m_ratio, muR_m_ratio, NLL, match_order, asinput);
  }
  double Resummation::DISxsNNLL(double x, double Q2, double y, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    DISy = y;
    return res_xs(sqrt(Q2), sqrt(Q2/x), muF_m_ratio, muR_m_ratio, NNLL, match_order, asinput);
  }
  double Resummation::DISxsNNNLL(double x, double Q2, double y, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    DISy = y;
    return res_xs(sqrt(Q2), sqrt(Q2/x), muF_m_ratio, muR_m_ratio, NNNLL, match_order, asinput);
  }

  double Resummation::xsLL   (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio,    LL, match_order, asinput);
  }
  double Resummation::xsNLL  (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio,   NLL, match_order, asinput);
  }
  double Resummation::xsNNLL (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio,  NNLL, match_order, asinput);
  }
  double Resummation::xsNNNLL(double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio, NNNLL, match_order, asinput);
  }


  double Resummation::xsRapLL   (double m, double Y, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs_rap(m, Y, sqrt_s, muF_m_ratio, muR_m_ratio,    LL, match_order, asinput);
  }
  double Resummation::xsRapNLL  (double m, double Y, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs_rap(m, Y, sqrt_s, muF_m_ratio, muR_m_ratio,   NLL, match_order, asinput);
  }
  double Resummation::xsRapNNLL (double m, double Y, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs_rap(m, Y, sqrt_s, muF_m_ratio, muR_m_ratio,  NNLL, match_order, asinput);
  }
  double Resummation::xsRapNNNLL(double m, double Y, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs_rap(m, Y, sqrt_s, muF_m_ratio, muR_m_ratio, NNNLL, match_order, asinput);
  }








  ////////////////////////////////
  //      NON-CLASS  PART       //
  ////////////////////////////////


  // for inclusive cross-section (invariant mass distribution)
  struct s_lum  { double x, mu, M; process p; collider c; PDFs *pdf; };
  double Lum_integrand(double z, void* par) {
    s_lum pars = *(s_lum *) par;
    double x   = pars.x;
    double mu  = pars.mu;
    double M   = pars.M;
    process p  = pars.p;
    collider c = pars.c;
    PDFs *pdf  = pars.pdf;
    double lum;
    if(p == ggH || p == ggA) lum = lumi::lumi_gg(pdf,z,x/z,mu)/z;
    else lum = lumi::lumi_DY(pdf,z,x/z,mu,M,p,c)/z;
    return lum;
  }
  double Lum_inclusive(PDFs *pdf, double x, double mu, double M, process p, collider c, int nf, double xTM, double y, double mc) {
    if(p < 0) { // DIS
      double TMtau = sqrt(1+4*xTM*xTM/M/M);
      return lumi::lumi_DIS(pdf,x,mu,nf,p,c,y,TMtau,mc/M);
    }
    double precision = 1e-4;
    s_lum pars = {x, mu, M, p, c, pdf};
    return integration::gauss(Lum_integrand,x,1.,precision,NULL,&pars);
  }
  double Lum_inclusive_rescaled(PDFs *pdf, double x, double mu, double M, process p, collider c, int nf, double xTM, double y, double mc) {
    //if(p < 0) { // DIS
    //  x *= 2./(1.+sqrt(1.+4.*TM*TM/M/M)); // target mass correction
    //}
    return Lum_inclusive(pdf, x, mu, M, p, c, nf, xTM, y, mc);
  }
  // for rapidity distributions
  double Lum_rapidity(PDFs *pdf, double x, double Y, double mu, double M, process p, collider c) {
    double x1 = sqrt(x) * exp(Y);
    double x2 = x/x1;  //sqrt(x) * exp(-Y);
    if(x1>=1 || x2>=1) return 0;
    double lum;
    if(p == ggH || p==ggA) lum = lumi::lumi_gg(pdf,x1,x2,mu);
    else lum = lumi::lumi_DY(pdf,x1,x2,mu,M,p,c);
    return lum;
  }
  double Lum_rapidity_rescaled(PDFs *pdf, double x, double Y, double mu, double M, process p, collider c) {
    x *= exp(-2*fabs(Y)); // rescaling!!
    return Lum_rapidity(pdf, x, Y, mu, M, p, c);
  }
  // for general use
  double Lum(double x, void *p) {
    Resummation *res = (Resummation *) p;
    if(res->proc < 0) { // Add TMCs
      double TMtau = sqrt(1+4*x*x*res->TargetMass*res->TargetMass/res->M/res->M);
      x *= 2./(1.+TMtau);
    }
    double lum=0;
    if(res->doRapidityDistr)
      lum = Lum_rapidity (res->pdfs, x, res->rapY, res->muF, res->M, res->proc, res->coll);
    else
      lum = Lum_inclusive(res->pdfs, x, res->muF, res->M, res->proc, res->coll, res->nf, res->tau*res->TargetMass, res->DISy, res->mass_c);
    return lum;
  }
  double LumRescaled(double x, void *p) {
    Resummation *res = (Resummation *) p;
    double lum=0;
    if(res->doRapidityDistr)
      lum = Lum_rapidity_rescaled (res->pdfs, x, res->rapY, res->muF, res->M, res->proc, res->coll);
    else
      lum = Lum_inclusive_rescaled(res->pdfs, x, res->muF, res->M, res->proc, res->coll, res->nf, res->tau*res->TargetMass, res->DISy, res->mass_c);
    return lum;
  }













  ////////////////////////////////
  //       PRIVATE  PART        //
  ////////////////////////////////

  void Resummation::InitMasses(double mc, double mb, double mt) {
    mass_c = mc;
    mass_b = mb;
    mass_t = mt;
  }
  void Resummation::InitMassThresholds(double mc, double mb, double mt) {
    mu_c = mc;
    mu_b = mb;
    mu_t = mt;
    /*
    cout << endl << "Using threshold masses from PDF set:" << endl
	 << "m_c = " << mu_c << " GeV" << endl
	 << "m_b = " << mu_b << " GeV" << endl
	 << "m_t = " << mu_t << " GeV" << endl
	 << endl;
    //*/
  }
  int Resummation::nf_of_mu(double mu) {
    //return 0;
    int res = 0;
    if(mu>=mu_t) res=6;
    else if(mu<mu_t && mu>=mu_b) res=5;
    //if(mu>=mu_b) res=5;
    else if(mu<mu_b && mu>=mu_c) res=4;
    else res=3;
    return res;
  }


  // Landau pole
  double Resummation::Nlandau() {
    double a = 0;
    if(APimpr == AP1) a = 1;
    if(APimpr == AP2) a = 2;
    if(APimpr == AP && (proc == ggH || proc == ggA)) a = 3;
    if(abs(prescr) != 10) a -= 0.5; // when prescr isn't MP00 or BP00
    //
    return exp(1./ab) - a;
  }




  void Resummation::ComputeCommon(bool redo_cheb) {
    nf = nf_of_mu(M); ///// muF, muR or m ?????????
    if(forceredo||!(last_nf==nf)) {
      b0 = QCD::b0(nf);
      b1 = QCD::b1(nf);
      b2 = QCD::b2(nf);
      b3 = QCD::b3(nf);
    }
    ab = 2*b0*as;
    if(proc < 0) ab = b0*as;  // DIS
    InitResummCoeffs();  // requires nf and m -> leave it here
    //
    if(redo_cheb) {
      luminosity_in_tau = Lum(tau, this);
      if(!firstinit) delete[] coeff;
      coeff = new double[Nchebyshev+1];
    }
  }





  dcomplex Resummation::mellD(int k, dcomplex N) {
    if(expDconst) {
      if(abs(prescr) == 1)
	return mell_D(k,N);
      else if(abs(prescr) == 2) {
	if(proc<0) return mell_Dbar(k,N); // DIS
	return mell_DB(k,N);
      }
      else if(abs(prescr) == 10)
	return mell_DMconst(k,N);
      else if(abs(prescr) == 11)
	return mell_DMconst(k,N-0.5);
    }
    else {
      if(abs(prescr) == 1)
	return mell_D_noconst(k,N);
      else if(abs(prescr) == 2) {
	if(proc<0) return mell_Dbar_noconst(k,N);  //DIS
	return mell_DB_noconst(k,N);
      }
      else if(abs(prescr) == 10)
	return mell_DM(k,N);
      else if(abs(prescr) == 11)
	return mell_DM(k,N-0.5);
    }
    return 0;
  }


  dcomplex Resummation::APimprovement(dcomplex (Resummation::*f)(dcomplex,void*), dcomplex N, void *p) {
    dcomplex res;
    if(proc == ggH || proc == ggA) {
      if(APimpr == AP0 ) res =   (this->*f)(N,p);
      if(APimpr == AP1 ) res =   (this->*f)(N+1,p);
      if(APimpr == AP2 ) res = 2*(this->*f)(N,p) - 3*(this->*f)(N+1,p) + 2*(this->*f)(N+2,p);
      if(APimpr == AP  ) res =   (this->*f)(N-1,p) - 2*(this->*f)(N,p) + 3*(this->*f)(N+1,p) - 2*(this->*f)(N+2,p) + (this->*f)(N+3,p);
    }
    else if(proc < 0) { // DIS
      if(prescr==MP2) {
	dcomplex Np = N+0.5+sqrt(N);
	dcomplex Nm = N+0.5-sqrt(N);
	if(APimpr == AP0 ) res = (this->*f)(Np,p)/2. + (this->*f)(Nm,p)/2.;
	if(APimpr == AP1 ) res = (this->*f)(Np+1,p)/2. + (this->*f)(Nm+1,p)/2.;
	if(APimpr == AP2 ) res = (this->*f)(Np,p)/4. + (this->*f)(Np+2,p)/4. + (this->*f)(Nm,p)/4. + (this->*f)(Nm+2,p)/4.;
	if(APimpr == AP  ) res = (this->*f)(Np,p)/4. + (this->*f)(Np+2,p)/4. + (this->*f)(Nm,p)/4. + (this->*f)(Nm+2,p)/4.;
      } else {
	if(APimpr == AP0 ) res = (this->*f)(N,p);
	if(APimpr == AP1 ) res = (this->*f)(N+1,p);
	if(APimpr == AP2 ) res = (this->*f)(N,p)/2. + (this->*f)(N+2,p)/2.;
	if(APimpr == AP  ) res = (this->*f)(N,p)/2. + (this->*f)(N+2,p)/2.;
      }
    }
    else { // DY
      if(APimpr == AP0 ) res = (this->*f)(N,p);
      if(APimpr == AP1 ) res = (this->*f)(N+1,p);
      if(APimpr == AP2 ) res = (this->*f)(N,p)/2. + (this->*f)(N+2,p)/2.;
      if(APimpr == AP  ) res = (this->*f)(N,p)/2. + (this->*f)(N+2,p)/2.;
    }
    return res;
  }


  dcomplex Resummation::NofZ(dcomplex z, double ab) {
    return exp(-z/ab);
  }
  dcomplex Resummation::ZofN(dcomplex N, double ab) {
    return -ab*log(N);
  }
  dcomplex Resummation::NofN(dcomplex N) {
    dcomplex NN = N;
    if(prescr == MP00) NN = N;      // original MP
    if(prescr == MP0 ) NN = exp(log(N-0.5));  // removes spurious sqrt(z)
    if(prescr == MP2 ) NN = exp(psi0(N));     // also restores more physical logs (correct analytic structure)
    if(prescr == MP5 ) NN = exp(2*psi0(N)-3*psi0(N+1.)+2*psi0(N+2.));     // reproduces PSI+AP2 of 1602.00695
    if(expBP && prescr>0) NN = N;
    return NN;
  }


  // Sudakov exponent
  dcomplex Resummation::S(dcomplex N, void *p) {
    dcomplex z = ZofN(N,ab);
    dcomplex res = g1(z)/as;
    if(LL_order > 0) res += g2(z);
    if(LL_order > 1) res += g3(z) *as;
    if(LL_order > 2) res += g4(z) *as*as;
    return res;
  }
  dcomplex Resummation::asS1(dcomplex N, void *p) {
    dcomplex L = ZofN(N,ab)/ab;
    dcomplex value;
    dcomplex L1 = L;
    dcomplex L2 = L*L;
    if(expBP) {
      L1 = mellD(0,N);
      L2 = 2*mellD(1,N) + 2*EulerGamma*L1;
    }
    if(g0constComputation) {
      L1 = -GD(1,1);
      L2 = -GD(2,1);
    }
    value = ab*L2*A1/b0 *ab/as/b0/2;
    if(LL_order > 0) value += ab*L1*(D1/2+2*A1*(log(M/muF)-EG))/b0;
    return value;
  }
  dcomplex Resummation::as2S2(dcomplex N, void *p) {
    dcomplex L = ZofN(N,ab)/ab;
    dcomplex value;
    dcomplex L1 = L;
    dcomplex L2 = L*L;
    dcomplex L3 = L2*L;
    if(expBP) {
      L1 = mellD(0,N);
      L2 = 2*mellD(1,N) + 2*EulerGamma*L1;
      L3 = 3*mellD(2,N)-3*ZETA2*L1 + 3*EulerGamma*L2 - 3*EulerGamma*EulerGamma*L1;
    }
    if(g0constComputation) {
      L1 = -GD(1,1);
      L2 = -GD(2,1);
      L3 = -GD(3,1);
    }
    value = - ab*ab*L3*A1/b0/3. *ab/as/b0/2;
    if(LL_order > 0) value += ab*ab*L2* ( A1*(EG-log(M/muR))/b0 + A2/2./b0/b0 -D1/4/b0);
    if(LL_order > 1) value += as*ab*L1* ( D2/2./b0 - 2.*A2*EG/b0 + D1*EG - 2.*A1*(EG*EG+Z2)
					  + (4.*A1*EG-D1)*log(M/muR) +2.*A2*log(M/muF)/b0
					  - 2.*A1*(pow(log(M/muR),2.)-pow(log(muF/muR),2.)) );
    return value;
  }
  dcomplex Resummation::as3S3(dcomplex N, void *p) {
    dcomplex L = ZofN(N,ab)/ab;
    dcomplex value;
    dcomplex L1 = L;
    dcomplex L2 = L*L;
    dcomplex L3 = L2*L;
    dcomplex L4 = L3*L;
    if(expBP) {
      L1 = mellD(0,N);
      L2 = 2*mellD(1,N) + 2*EulerGamma*L1;
      L3 = 3*mellD(2,N)-3*ZETA2*L1 + 3*EulerGamma*L2 - 3*EulerGamma*EulerGamma*L1;
      L4 = 4*mellD(3,N)-12*ZETA2*mellD(1,N)+8*ZETA3*L1 + 4*EulerGamma*L3 - 6*EulerGamma*EulerGamma*L2 + 4*EulerGamma*EulerGamma*EulerGamma*L1;
    }
    if(g0constComputation) {
      L1 = -GD(1,1);
      L2 = -GD(2,1);
      L3 = -GD(3,1);
      L4 = -GD(4,1);
    }
    value = ab*ab*ab*L4*A1/b0/6. *ab/as/b0/2;
    if(LL_order > 0) value += -ab*ab*ab*L3/b0/b0/6.* (2*A2 - b0*D1 + A1*(b1 + 4*b0*(EG-log(M/muR))) );
    if(LL_order > 1) value += as*ab*ab*L2/b0/b0/2* ( A3 + b0*(-D2 + 4*A2*EG - b1*D1/2 -2*b0*EG*D1
							      + 2.*A1*(b1*EG + 2*b0*(EG*EG+Z2)))
						     - 2*b0*(2*A2 - b0*D1 + A1*(b1+4*b0*EG))*log(M/muR)
						     + 4*A1*b0*b0*pow(log(M/muR),2.) );
    if(LL_order > 2) value += as*as*ab*L1/b0/6* (3*D3 - 12*A3*EG + 6*b0*b1*D1*EG + 12*b0*D2*EG - 24*A2*b0*pow(EG,2)
						 - 12*A1*b0*b1*pow(EG,2) + 12*pow(b0,2)*D1*pow(EG,2) - 16*A1*pow(b0,2)*pow(EG,3) - 24*A2*b0*Z2
						 - 12*A1*b0*b1*Z2 + 12*pow(b0,2)*D1*Z2 - 48*A1*pow(b0,2)*EG*Z2 - 32*A1*pow(b0,2)*Z3
						 + 6*(2*A3 + b0*(-(b1*D1) - 2*D2 + 8*A2*EG + 4*A1*b1*EG - 4*b0*D1*EG + 8*A1*b0*pow(EG,2) + 8*A1*b0*Z2))*log(M/muR)
						 - 12*b0*(2*A2 + A1*b1 - b0*D1 + 4*A1*b0*EG)*pow(log(M/muR),2)
						 + 16*A1*pow(b0,2)*pow(log(M/muR),3) - 12*A3*log(muF/muR)
						 + 12*b0*(2*A2 + A1*b1)*pow(log(muF/muR),2) - 16*A1*pow(b0,2)*pow(log(muF/muR),3));
    return value;
  }
  //
  // constants
  double Resummation::g0() {
    double value = 1.;
    if(g0exp) {
      if(LL_order > 0+LogarithmicCounting) value = exp(g01*as);
      if(LL_order > 1+LogarithmicCounting) value = exp(g01*as + (g02-g01*g01/2.)*as*as);
      if(LL_order > 2+LogarithmicCounting) value = exp(g01*as + (g02-g01*g01/2.)*as*as + (g03-g01*g02+g01*g01*g01/3.)*as*as*as);
    } else if(expDconst && prescr<0) { // psi-soft with same constants as A-soft
	// construct pert expansion of g0/g0bar
	double g0rat1 = g01 - g0bar1;
	double g0rat2 = g02 - g01*g0bar1 + g0bar1*g0bar1 - g0bar2;
	double g0rat3 = g03 - g02*g0bar1 + g01*g0bar1*g0bar1 - g0bar1*g0bar1*g0bar1 - g01*g0bar2 + 2*g0bar1*g0bar2 - g0bar3;
	if(LL_order > 0+LogarithmicCounting) value = (1+as*g0bar1) * exp(g0rat1*as);
	if(LL_order > 1+LogarithmicCounting) value = (1+as*g0bar1 + as*as*g0bar2) * exp(g0rat1*as + (g0rat2-g0rat1*g0rat1/2.)*as*as);
	if(LL_order > 2+LogarithmicCounting) value = (1+as*g0bar1 + as*as*g0bar2 + as*as*as*g0bar3) * exp(g0rat1*as + (g0rat2-g0rat1*g0rat1/2.)*as*as + (g0rat3-g0rat1*g0rat2+g0rat1*g0rat1*g0rat1/3.)*as*as*as);
    } else {
      if(LL_order > 0+LogarithmicCounting) value += g01 * as;
      if(LL_order > 1+LogarithmicCounting) value += g02 * as*as;
      if(LL_order > 2+LogarithmicCounting) value += g03 * as*as*as;
    }
    return value;
  }
  double Resummation::g0sub() {
    double value = 0;
    if(LL_order > 0+LogarithmicCounting) value += g0sub1 * as;
    if(LL_order > 1+LogarithmicCounting) value += g0sub2 * as*as;
    if(LL_order > 2+LogarithmicCounting) value += g0sub3 * as*as*as;
    return value;
  }
  //
  double Resummation::asg01() {
    if(LL_order > 0+LogarithmicCounting) return g01*as;
    return 0;
  }
  double Resummation::as2g02() {
    if(LL_order > 1+LogarithmicCounting) return g02*as*as;
    if(g0exp) {    
      if(LL_order > 0+LogarithmicCounting) return g01*g01*as*as/2.;
    } else if(expDconst && prescr<0) { // psi-soft with same constants as A-soft
      if(LL_order > 0+LogarithmicCounting) return (g01*g01-g0bar1*g0bar1)*as*as/2.;
    }
    return 0;
  }
  double Resummation::as3g03() {
    if(LL_order > 2+LogarithmicCounting) return g03*as*as*as;
    if(g0exp) {    
      if(LL_order > 1+LogarithmicCounting) return (g02 - g01*g01/3.)*g01*as*as*as;
      if(LL_order > 0+LogarithmicCounting) return g01*g01*g01*as*as*as/6.;
    } else if(expDconst && prescr<0) { // psi-soft with same constants as A-soft
      if(LL_order > 1+LogarithmicCounting) return (-g01*g01*g01 + 3*g01*g02 + g0bar1*g0bar1*g0bar1 - 3*g0bar1*g0bar2)/3.*as*as*as;
      if(LL_order > 0+LogarithmicCounting) return (g01*g01*g01 - 3*g01*g0bar1*g0bar1 + 2*g0bar1*g0bar1*g0bar1)/6.*as*as*as;
    }
    return 0;
  }
  //
  //
  dcomplex Resummation::expS(dcomplex N) {
    if(expBP) return exp( BorelPrescriptionExp(N) );
    return exp( APimprovement(&Resummation::S,N) );
  }
  dcomplex Resummation::g0expS_order1(dcomplex N) {
    dcomplex value;
    value = APimprovement(&Resummation::asS1,N)
      + asg01();
    return value;
  }
  dcomplex Resummation::g0expS_order2(dcomplex N) {
    dcomplex value;
    value = pow(APimprovement(&Resummation::asS1,N),2.)/2. + APimprovement(&Resummation::as2S2,N)
      + asg01() * APimprovement(&Resummation::asS1,N)
      + as2g02();
    return value;
  }
  dcomplex Resummation::g0expS_order3(dcomplex N) {
    dcomplex value;
    value = pow(APimprovement(&Resummation::asS1,N),3.)/6. + APimprovement(&Resummation::asS1,N)*APimprovement(&Resummation::as2S2,N) + APimprovement(&Resummation::as3S3,N)
      + asg01()*( pow(APimprovement(&Resummation::asS1,N),2.)/2. + APimprovement(&Resummation::as2S2,N) )
      + as2g02()*APimprovement(&Resummation::asS1,N)
      + as3g03();
    return value;
  }
  //
  dcomplex Resummation::g0expS_matched_predamping(dcomplex N) {
    dcomplex value;
    //
    as/=asrescale; ab/=asrescale;
    value = expS(N) * g0() - g0sub() - 1.;
    //
    if(matching_order > 0) value -= g0expS_order1(N) - (LL_order>0+LogarithmicCounting ? g0sub1*as : 0);
    if(matching_order > 1) value -= g0expS_order2(N) - (LL_order>1+LogarithmicCounting ? g0sub2*as*as : 0);
    if(matching_order > 2) value -= g0expS_order3(N) - (LL_order>2+LogarithmicCounting ? g0sub3*as*as*as :0);
    //
    as*=asrescale; ab*=asrescale;
    value*=pow(asrescale,matching_order+1);
    //
    return value;
  }
  dcomplex Resummation::g0expS_matched_presubtraction(dcomplex N) {
    if(DampingOrder > 0) {
      dcomplex res = 0;
      for(unsigned int i=1; i<=DampingOrder; i++) {
	res += pow(-1.,i+1) * gsl_sf_choose(DampingOrder,i) * g0expS_matched_predamping(N+i);
      }
      return res;
    }
    return g0expS_matched_predamping(N);
  }
  dcomplex Resummation::g0expS_matched(dcomplex N) {
    if(doNsubtraction) {
      return g0expS_matched_presubtraction(N) - g0expS_matched_presubtraction(Nsubtraction);
    }
    return g0expS_matched_presubtraction(N);
  }







  ///////////////////////////
  //  BOREL  PRESCRIPTION  //
  ///////////////////////////

  dcomplex Resummation::cheb_term(dcomplex xi, int j) {
    dcomplex val;
    if     (prescr == BP00) val = TruncatedGamma(xi+j+1,   log(1./tau));
    else if(prescr == BP0 ) val = TruncatedGamma(xi+j+1, 2*log(1./tau)) * exp((xi+j+1)*log(2.));
    else if(prescr == BP1 ) val = exp((xi+j+1.)*log(1.-tau))/(xi+j+1.);
    else if(prescr == BP2 ) val = IncompleteBeta(xi+j+1, 1.-xi/2., 1.-tau);
    if(proc<0 && prescr == BP2 ) val = IncompleteBeta(xi+j+1, 1.-xi, 1.-tau); // DIS // check this!!
    return val;
  }
  dcomplex Resummation::delta_term(dcomplex xi) {
    dcomplex val = cheb_term(xi,-1);
    if(expDconst) val -= 1./xi;
    return val;
  }
  dcomplex Resummation::convolution(dcomplex xi) { 
    dcomplex value = 0.;
    for(unsigned int j=0; j<=Nchebyshev; j++) {
      value += coeff[j] * cheb_term(xi,j);
    }
    //return value + luminosity_in_tau * delta_term(xi);
    return value + delta_term(xi);
  }
  dcomplex Resummation::BorelIntegrand(double w, dcomplex xi) {
    //dcomplex result = g0expS_matched(w/xi)*exp(-w/ab) * convolution(xi) /Gamma(xi+1)/ab; // /2./M_PI
    dcomplex N = NofZ(w/xi,ab);
    dcomplex result = g0expS_matched(N)*exp(-w/ab) * convolution(xi) /Gamma(xi+1)/ab; // /2./M_PI
    return result;
  }
  double Resummation::BorelIntegrand_chvar(double x1, double x2) {
    double w = x1*BPcutoff;
    //
    double NL = Nlandau();
    //double cutlength = BPcutoff/ab/log(NL);
    double cutlength = w/ab/log(NL);
    //
    double r = 1.3 + 2*cutlength;
    dcomplex xi = r * exp(I*x2*M_PI);
    double result = real(xi*BorelIntegrand(w, xi)) * BPcutoff;  // *2.*M_PI;
    //double r = 1. + 2*cutlength;
    //dcomplex xi = r * exp(I*x2*M_PI) - 0.5*cutlength;
    //double result = real((xi+0.5*cutlength)*BorelIntegrand(w, xi)) * BPcutoff;  // *2.*M_PI;
    //
    //cout << result << endl;
    return result;
  }
  int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    Resummation *Bres = (Resummation *) pars;
    res[0] = Bres->BorelIntegrand_chvar(x[0],x[1]);
    return 0;
  }
  double Resummation::BorelPrescription() {
    double cres[1], error[1], prob[1];
    double epsrel=1.e-3, epsabs=1.e-100;
    int last = 4;
    int verbose = 0;
    int nregions, neval, fail;
    Cuhre(2, 1, integrandCUBA, this, 1,
	  epsrel, epsabs, verbose | last,
	  0, 500000, 9, NULL, NULL,
	  &nregions, &neval, &fail, cres, error, prob);
    if(fail == -1) {
      cout << "  ERROR: dimension out of range." << endl;
    } else if(fail>0) {
      cout << "  ERROR: precision not reached." << endl;
    }
    return cres[0];
  }






  ////////////////////////////////////////
  // Borel Prescription at the exponent //
  ////////////////////////////////////////

  dcomplex Resummation::NLogGen(dcomplex N, void *par) {
    dcomplex res;
    dcomplex xi = *(dcomplex *)par;
    if     (prescr == BP00) res = exp(-xi*log(N));
    else if(prescr == BP0 ) res = exp(-xi*log(N-0.5));
    else if(prescr == BP1 ) res = Gamma(N)/Gamma(N+xi);
    else if(prescr == BP2 ) res = Gamma(N-xi/2.)/Gamma(N+xi/2.);
    if(proc<0 && prescr == BP2 ) res = Gamma(N-xi)/Gamma(N);
    if(expDconst) res -= 1./Gamma(1+xi);
    //if(expDconst) res -= Gamma(1-xi/2.)/Gamma(1+xi/2.);
    return res;
  }
  dcomplex Resummation::BorelIntegrandExp(double w, dcomplex xi, dcomplex N) {
    dcomplex result = S(NofZ(w/xi,ab),NULL)*exp(-w/ab) * APimprovement(&Resummation::NLogGen,N,&xi) /xi/ab; // /2./M_PI
    return result;
  }
  dcomplex Resummation::BorelIntegrandExp_chvar(double x1, double x2, dcomplex N) {
    double w = x1*BPcutoff;
    //
    double NL = Nlandau();
    //double cutlength = BPcutoff/ab/log(NL);
    double cutlength = w/ab/log(NL);
    //
    double r = 1.3 + 2*cutlength;
    //double r = 0.7 + 2*cutlength;
    //double r = 1.3 + 1.5*cutlength;
    //double r = 1.5*cutlength;
    dcomplex xi = r * exp(I*x2*2*M_PI);
    dcomplex result = xi*BorelIntegrandExp(w, xi, N) * BPcutoff;  // *2.*M_PI;
    //double r = 1. + 2*cutlength;
    //dcomplex xi = r * exp(I*x2*2*M_PI) - 0.5*cutlength;
    //double result = real((xi+0.5*cutlength)*BorelIntegrand(w, xi)) * BPcutoff;  // *2.*M_PI;
    //
    //cout << result << endl;
    return result;
  }
  struct BPexpintpars {
    Resummation *res;
    dcomplex N;
  };
  int integrandCUBAexp(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    BPexpintpars p = *(BPexpintpars *) pars;
    Resummation *Bres = p.res;
    dcomplex N = p.N;
    dcomplex r = Bres->BorelIntegrandExp_chvar(x[0],x[1],N);
    res[0] = real(r);
    res[1] = imag(r);
    return 0;
  }
  dcomplex Resummation::BorelPrescriptionExp(dcomplex N) {
    double cres[2], error[2], prob[2];
    double epsrel=2.e-2, epsabs=1.e-100;
    //double epsrel=5.e-3, epsabs=1.e-100;
    int last = 4;
    int verbose = 0;
    int nregions, neval, fail;
    nregions=1; // to remove the warning
    int mineval = 0, maxeval = 40000000;
    BPexpintpars pars = { this, N };
    Cuhre(2, 2, integrandCUBAexp, &pars, 1, epsrel, epsabs, verbose | last, mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    //Suave(2, 2, integrandCUBAexp, &pars, epsrel, epsabs, verbose | last, 0, mineval, maxeval, 1000, 25., &nregions, &neval, &fail, cres, error, prob);
    //Vegas(2, 2, integrandCUBAexp, &pars, epsrel, epsabs, verbose | last, 0, mineval, maxeval, 10000, 2000, 10000, 1, NULL, &neval, &fail, cres, error, prob);
    if(fail == -1) {
      cout << "  ERROR: dimension out of range." << endl;
    } else if(fail>0) {
      cout << "  ERROR: precision not reached." << endl;
    }
    return dcomplex(cres[0],cres[1]);
  }






  ///////////////////////////
  //  MINIMAL PRESCRIPTION //
  ///////////////////////////

  dcomplex Resummation::LuminosityN(dcomplex N) {
    dcomplex value = 0;
    /*
    if(0&&doRapidityDistr) { // apparently I'm already using a luminosity of the form of the integrated one, so I must not re-introduce rapidity here
      for(unsigned int i=0; i<=Nchebyshev; i++) {
	for(int j=0; j<=MPgamma; j++) {
	  for(int k=0; k<=i; k++) {
	    value += coeff[i] * pow(-1.,j) * gsl_sf_choose(MPgamma,j) * pow(2*abs(rapY),i-k) /gsl_sf_fact(i-k) / pow(N+j-MPbeta,k+1.);
	  }
	}
      }
      value *= exp(-2*N*abs(rapY));
    } else {
    */
    for(unsigned int i=0; i<=Nchebyshev; i++) {
      for(int j=0; j<=MPgamma; j++) {
	value += coeff[i] * pow(-1.,j) * gsl_sf_choose(MPgamma,j) / pow(N+j-MPbeta,i+1.);
      }
    }
    if(doRapidityDistr) value *= exp(-2*N*fabs(rapY));
    //if(proc<0) value *= exp(N*log(2./(1.+sqrt(1.+4.*TargetMass*TargetMass/M/M))));
    //
    return value;
  }
  dcomplex Resummation::lum_times_g0expS_matched(dcomplex N) {
    return g0expS_matched(NofN(N)) * LuminosityN(N);
  }
  double Minimal_integrand(double t, void *p) {
    Resummation *Mres = (Resummation *) p;
    double x = Mres->tau;
    if(Mres->proc < 0) {
      double tau = sqrt(1+4*x*x*Mres->TargetMass*Mres->TargetMass/Mres->M/Mres->M);
      x *= 2./(1.+tau);
      if(Mres->proc<=-6 && Mres->proc>=-9) {  // CC charm production: FFNS
	x *= 1.+pow(Mres->mass_c/Mres->M,2);
      }
    }
    dcomplex N = Mres->MP_reN0 + (Mres->MP_slope-I)*log(t);
    dcomplex res = (I-Mres->MP_slope) * exp(-N*log(x)) * Mres->lum_times_g0expS_matched(N)/t;
    return imag(res)/M_PI;
  }
  /*
  double Minimal_integrand_inf(double t, void *p) {
    Resummation Mres = *(Resummation *) p;
    double x = Mres.tau;
    dcomplex N = Mres.MP_reN0 + (I-Mres.MP_slope)*t;
    dcomplex res = (I-Mres.MP_slope) * exp(-N*log(x)) * Mres.lum_times_g0expS_matched(N);
    return imag(res)/M_PI;
  }
  */
  dcomplex Resummation::InverseMellinIntegrand(dcomplex N) {
    double x = tau;
    if(proc < 0) {
      double TMtau = sqrt(1+4*x*x*TargetMass*TargetMass/M/M);
      x *= 2./(1.+TMtau);
    }
    return exp( -N*log(x) ) * lum_times_g0expS_matched(N) / M_PI;
  }
  double Resummation::InverseMellin() { // Talbot algorithm
    if(tau==1.) return 0;
    //double log1ox = -log(tau);
    double res = 0.; 
    // prec ~ 10^{-0.6*MM}
    int MM = 18; // tune
    //double r = 1.+2.*MM/5./log1ox;
    double r = 2.5; // tune
    double t, cott, sigma;
    dcomplex N;
    res = real(0.5 * InverseMellinIntegrand(r));
    for(int k=1; k<MM; k++) {
      t = k*M_PI/MM;
      cott = 1./tan(t);
      sigma = t + cott*(t*cott - 1.);
      N = r*t*(cott+I);
      res += real(InverseMellinIntegrand(N) * (1.+I*sigma));
    }
    res *= M_PI*r/MM;
    return res;
  }
  double Resummation::MinimalPrescription() {
    //return InverseMellin();
    double res=0, prec = 1.e-3;
    if(expBP) prec *= 5;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(Minimal_integrand,0.,1.,prec,NULL,this);
    //res = integration::gauss_inf(Minimal_integrand_inf,0.,prec,NULL,&x);  // fails..
    gsl_set_error_handler (old_handler);
    return res;
  }

  /*
  double rap_Minimal(double Y, double x) {
    em2Y = exp(-2*ABS(Y));
    double xe2Y = x/em2Y;
    double prec = 1e-4;
    //return integration::dgauss(Minimal_integrand,0.,1.,prec,NULL,&xe2Y);
    return integration::gauss_inf(Minimal_integrand_inf,0.,prec,NULL,&xe2Y);  // faster
  }
  */










  ///////////////////////////
  //  AUXILIARY FUNCTIONS  //
  ///////////////////////////



  // functions in the exponent; gk means N^(k-1)LL approximation
  // z is the variable ab*log(1/N) = 2 as b0 log(1/N)
  dcomplex Resummation::g1(dcomplex z) {
    return (log(z+1)*(z+1)-z)*A1/b0/b0;
  }
  dcomplex Resummation::g2(dcomplex z) {
    return (( pow(log(z+1),2)*0.5 - z + log(z+1) ) * ( A1*b1/b0/b0 ) +
	    (z - log(z+1) ) * A2/b0/b0
	    - (log(z+1) * (EG - log(M/muR)) + z*log(muF/muR) )*2.*A1/b0
	    + D1/2./b0*log(1+z));
  }
  // g3 taken from Moch-Vermaseren-Vogt, hep-ph/0506288
  dcomplex Resummation::g3(dcomplex z) {
    return ( (z*z/(z+1.)) * (A3 - A1*b2 + A1*b1*b1 - A2*b1)/2./b0/b0/b0
	     + ( log(z+1.)/(z+1.) + pow(log(z+1.),2.)/(z+1.)/2. ) * A1*b1*b1/b0/b0/b0
	     + log(z+1.) * A1*(b2-b1*b1)/b0/b0/b0
	     + ( z/(z+1.) - log(z+1.)/(z+1.) ) * ( 2.*A1*b1*EG + A2*b1/b0 - D1/2.*b1 )/b0/b0
	     - ( z/(z+1.) ) * ( A1*b2/b0/b0 + 2.*A1*(EG*EG + Z2) + 2.*A2*EG/b0 )/b0
	     + ( z/(z+1.) ) * ( D2/2./b0/b0 + D1*EG/b0 )
	     + ( ( z/(z+1.) ) * ( 2.*A1*EG + A2/b0 -A1*b1/b0 -D1/2. )/b0
	      + log(z+1.)/(z+1.) * A1*b1/b0/b0 ) * 2.*log(M/muR)
	     - z * A2/b0/b0 * 2.*log(muF/muR)
	     + ( z * pow(log(muF/muR),2.) - z/(z+1.) * pow(log(M/muR),2.) ) * 2.*A1/b0 
	     )*b0;
  }
  //
  // g4 taken from Moch-Vermaseren-Vogt, hep-ph/0506288 (translated from tex to C++ with Mathematica)
  template<class T> T Power(T x, unsigned int n) {
    T res = 1;
    for(unsigned int i=0; i<n; i++) res *= x;
    return res;
  }
  dcomplex Resummation::g4(dcomplex z) {
    return -(-6*A3*b1*z + 6*A2*Power(b1,2)*z - 6*A1*b1*b2*z + 6*A1*b3*z + 6*b0*b1*D2*z - 6*b0*D3*z + 
      24*A3*b0*EG*z - 24*A2*b0*b1*EG*z - 24*Power(b0,2)*D2*EG*z + 
      48*A2*Power(b0,2)*Power(EG,2)*z - 24*Power(b0,3)*D1*Power(EG,2)*z + 
      32*A1*Power(b0,3)*Power(EG,3)*z - 6*A4*Power(z,2) + 3*A3*b1*Power(z,2) + 
      3*A2*Power(b1,2)*Power(z,2) - 9*A1*b1*b2*Power(z,2) + 9*A1*b3*Power(z,2) - 
      3*b0*Power(b1,2)*D1*Power(z,2) + 3*b0*b2*D1*Power(z,2) + 3*b0*b1*D2*Power(z,2) - 
      3*b0*D3*Power(z,2) + 12*A3*b0*EG*Power(z,2) - 12*A2*b0*b1*EG*Power(z,2) + 
      12*A1*b0*Power(b1,2)*EG*Power(z,2) - 12*A1*b0*b2*EG*Power(z,2) - 
      12*Power(b0,2)*D2*EG*Power(z,2) + 24*A2*Power(b0,2)*Power(EG,2)*Power(z,2) - 
      12*Power(b0,3)*D1*Power(EG,2)*Power(z,2) + 
      16*A1*Power(b0,3)*Power(EG,3)*Power(z,2) - 4*A4*Power(z,3) + 4*A3*b1*Power(z,3) - 
      4*A2*Power(b1,2)*Power(z,3) + 4*A1*Power(b1,3)*Power(z,3) + 4*A2*b2*Power(z,3) - 
      8*A1*b1*b2*Power(z,3) + 4*A1*b3*Power(z,3) + 48*A2*Power(b0,2)*z*Z2 - 
      24*Power(b0,3)*D1*z*Z2 + 96*A1*Power(b0,3)*EG*z*Z2 + 
      24*A2*Power(b0,2)*Power(z,2)*Z2 - 12*Power(b0,3)*D1*Power(z,2)*Z2 + 
      48*A1*Power(b0,3)*EG*Power(z,2)*Z2 + 64*A1*Power(b0,3)*z*Z3 + 
      32*A1*Power(b0,3)*Power(z,2)*Z3 - 16*A1*Power(b0,3)*z*(2 + z)*Power(log(M/muR),3) + 
      24*A3*b0*z*Power(1 + z,2)*log(muF/muR) - 
      24*Power(b0,2)*(2*A2 + A1*b1)*z*Power(1 + z,2)*Power(log(muF/muR),2) + 
      32*A1*Power(b0,3)*z*Power(1 + z,2)*Power(log(muF/muR),3) + 6*A3*b1*log(1 + z) - 
      6*A2*Power(b1,2)*log(1 + z) + 6*A1*b1*b2*log(1 + z) - 6*A1*b3*log(1 + z) - 
      6*b0*b1*D2*log(1 + z) + 24*A2*b0*b1*EG*log(1 + z) - 
      12*Power(b0,2)*b1*D1*EG*log(1 + z) + 
      24*A1*Power(b0,2)*b1*Power(EG,2)*log(1 + z) + 12*A1*b1*b2*z*log(1 + z) - 
      12*A1*b3*z*log(1 + z) - 6*A1*Power(b1,3)*Power(z,2)*log(1 + z) + 
      12*A1*b1*b2*Power(z,2)*log(1 + z) - 6*A1*b3*Power(z,2)*log(1 + z) + 
      24*A1*Power(b0,2)*b1*Z2*log(1 + z) - 6*A2*Power(b1,2)*Power(log(1 + z),2) + 
      3*b0*Power(b1,2)*D1*Power(log(1 + z),2) - 12*A1*b0*Power(b1,2)*EG*Power(log(1 + z),2) + 
      2*A1*Power(b1,3)*Power(log(1 + z),3) + 
      12*Power(b0,2)*Power(log(M/muR),2)*((2*A2 - b0*D1 + 4*A1*b0*EG)*z*(2 + z) + 
         2*A1*b1*log(1 + z)) - 12*b0*log(M/muR)*
       (z*(-2*b0*D2 - 4*Power(b0,2)*D1*EG + 8*A1*Power(b0,2)*Power(EG,2) + 
            A1*Power(b1,2)*z - A1*b2*z - b0*D2*z - 2*Power(b0,2)*D1*EG*z + 
            4*A1*Power(b0,2)*Power(EG,2)*z + A3*(2 + z) - A2*(b1 - 4*b0*EG)*(2 + z) + 
            8*A1*Power(b0,2)*Z2 + 4*A1*Power(b0,2)*z*Z2) + 
         b1*(2*A2 - b0*D1 + 4*A1*b0*EG)*log(1 + z) - A1*Power(b1,2)*Power(log(1 + z),2)))/
   (12.*Power(b0,2)*Power(1 + z,2));
  }



  void Resummation::InitResummCoeffs() {
    if(forceredo||last_nf!=nf) InitCusp();
    if(proc == ggH) {  // Higgs
      if(forceredo||!(last_nf==nf)) InitCoeffHiggs();
      InitConstHiggs(false);
    } else if(proc == ggA) {  // Pseudoscalar
      if(forceredo||!(last_nf==nf)) InitCoeffHiggs();
      InitConstHiggs(true);
    } else if(proc > 0) {  // Drell-Yan
      if(forceredo||!(last_nf==nf)) InitCoeffDY();
      InitConstDY();
    } else if(proc < 0) {  // DIS
      if(forceredo||!(last_nf==nf)) InitCoeffDIS();
      InitConstDIS();
    }
    // cout << "g0bar1 = " << g0bar1 << endl;
    // cout << "g0bar2 = " << g0bar2 << endl;
    // cout << "g0bar3 = " << g0bar3 << endl;
    // cout << "g01    = " << g01 << endl;
    // cout << "g02    = " << g02 << endl;
    // cout << "g03    = " << g03 << endl;
    return;
  }

  void Resummation::InitCusp() {
    using namespace QCD;
    A1 = CF/M_PI;
    A2 = CF/2.* (67.*CA/18.-5./9.*nf-CA*ZETA2) /M_PI/M_PI;
    A3 = ( CA*CA* (245./24.-67.*ZETA2/9.+11.*ZETA3/6.+11.*ZETA2*ZETA2/5.) +
	   ( (2.*ZETA3 - 55./24.)*CF + (10.*ZETA2/9.-7.*ZETA3/3.-209./108.)*CA )*nf
	   - nf*nf*1./27. ) *CF/4. /M_PI/M_PI/M_PI;
    A4 = 0; // use Pade' [1/1]
    if(nf == 3) A4 = 0.314757;
    if(nf == 4) A4 = 0.172958;
    if(nf == 5) A4 = 0.0622776;
  }
  void Resummation::InitCoeffDY() {
    EG = EulerGamma;
    Z2 = ZETA2;
    Z3 = ZETA3;
    using namespace QCD;
    D1 = 0;
    D2 = CF * ( CA * (176.*ZETA2/3. + 56.*ZETA3 - 1616./27.) + nf*(224./27.-32.*ZETA2/3.) )/16./M_PI/M_PI; // checked: OK.
    D3 = CF * ( CA*CA *(-594058./729. +98224./81.*ZETA2 +40144./27.*ZETA3 -2992./15.*ZETA2*ZETA2 -352./3.*ZETA2*ZETA3 -384.*ZETA5 )
		+CA*nf*( 125252./729. -29392./81.*ZETA2 -2480./9.*ZETA3 +736./15.*ZETA2*ZETA2 )
		+CF*nf*( 3422./27. - 32.*ZETA2 - 608./9.*ZETA3 - 64./5.*ZETA2*ZETA2 )
		+nf*nf*( 320./27.*ZETA3 + 640./27.*ZETA2 - 3712./729.)) /64./M_PI/M_PI/M_PI; // checked: OK.
    return;
  }
  void Resummation::InitCoeffHiggs() {
    using namespace QCD;
    InitCusp();
    InitCoeffDY();
    A1 *= CA/CF;
    A2 *= CA/CF;
    A3 *= CA/CF;
    A4 *= CA/CF;
    D2 *= CA/CF;
    D3 *= CA/CF;
    //
    // prints g4 at order lambda (I used it for ggHiggs, for the NNLL curve)
    // cout << endl << "b0: " << b0*4*M_PI;
    // cout << endl << "b1: " << beta1*4*M_PI*4*M_PI;
    // cout << endl << "A1: " << A1*4*M_PI;
    // cout << endl << "A2: " << A2*4*M_PI*4*M_PI;
    // cout << endl << "A3: " << A3*4*M_PI*4*M_PI*4*M_PI;
    // cout << endl << "D2: " << D2*4*M_PI*4*M_PI;
    // cout << endl << "D3: " << setprecision(7) << D3*4*M_PI*4*M_PI*4*M_PI;
    // cout << endl << "g:  " << EulerGamma;
    // cout << endl << "z2: " << ZETA2;
    // cout << endl << "z3: " << ZETA3;
    // cout << endl << -D3 + (4*A3 - 4*D2*b0)*EulerGamma  + (8*A2*b0 + 4*A1*b1*b0) * (EulerGamma*EulerGamma + ZETA2)
    //   + 16./3.*A1*b0*b0* (3*EulerGamma*ZETA2 + EulerGamma*EulerGamma*EulerGamma + 2*ZETA3) << endl;
    // b0 *= 4*M_PI;
    // b1 *= 4*M_PI*4*M_PI;
    // A1 *= 4*M_PI;
    // A2 *= 4*M_PI*4*M_PI;
    // A3 *= 4*M_PI*4*M_PI*4*M_PI;
    // D2 *= 4*M_PI*4*M_PI;
    // D3 *= 4*M_PI*4*M_PI*4*M_PI;
    // double res = -D3 + (4*A3 - 4*D2*b0)*EulerGamma  + (8*A2*b0 + 4*A1*b1*b0) * (EulerGamma*EulerGamma + ZETA2)
    //   + 16./3.*A1*b0*b0* (3*EulerGamma*ZETA2 + EulerGamma*EulerGamma*EulerGamma + 2*ZETA3);
    // cout << endl << res << endl << res/pow(4*M_PI,3) << endl;
    return;
  }
  void Resummation::InitCoeffDIS() {
    EG = EulerGamma/2;
    Z2 = ZETA2/4;
    Z3 = ZETA3/8;
    using namespace QCD;
    double B1 = -3*CF;
    double B2 = CF*CF*(-3./2 + 12*ZETA2 - 24*ZETA3) +
      CF*CA*(-3155./54 + 44./3*ZETA2 + 40.*ZETA3) +
      CF*nf*(247./27 - 8./3*ZETA2);
    double B3 = CF*CF*CF* (-29./2 - 18*ZETA2 - 68*ZETA3 - 288./5*ZETA2*ZETA2 +
			   32*ZETA2*ZETA3 + 240*ZETA5) +
      CA*CF*CF*(-46. + 287.*ZETA2 - 712./3*ZETA3 - 272./5*ZETA2*ZETA2 -
		16*ZETA2*ZETA3 - 120*ZETA5) +
      CA*CA*CF*(-599375./729 + 32126./81*ZETA2 + 21032./27*ZETA3 -
		652./15*ZETA2*ZETA2 - 176./3*ZETA2*ZETA3 - 232.*ZETA5) +
      CF*CF*nf*(5501./54 - 50.*ZETA2 + 32./9*ZETA3) +
      CF*nf*nf*(-8714./729 + 232./27*ZETA2 - 32./27*ZETA3) +
      CA*CF*nf*(160906./729 - 9920./81*ZETA2 - 776./9*ZETA3 +
		208./15*ZETA2*ZETA2);
    D1 = 2*B1/4/M_PI;
    D2 = 2*B2/4/M_PI/4/M_PI;
    D3 = 2*B3/4/M_PI/4/M_PI/4/M_PI;
    return;
  }

  void Resummation::InitConstDY() {
    using namespace QCD;
    //
    // coefficients of the delta(1-z) at central scales
    if(forceredo||!(last_nf==nf && last_M==M && last_mtop==Hmtop && last_EFT==mtEFT)) {
      d1c = CF*(2*ZETA2-4);
      d2c =  ( CA*CF * ( - 12./5. * ZETA2*ZETA2 + 592./9. * ZETA2 + 28. * ZETA3 - 1535./12. )
	       + CF*CF * ( 8./5. * ZETA2*ZETA2 - 70. * ZETA2 - 60. * ZETA3 + 511./4. )
	       + nf*CF * ( 8. * ZETA3 - 112./9. * ZETA2 + 127./6. ) ) /16.;
      // Cth3 and equation taken from 1405.4827
      double NFV = 0; // ?????
      double Cth3 = CF*CF*CF*(-5599./384. -65./96.*ZETA2 -153./80.*ZETA2*ZETA2 +803./630.*pow(ZETA2,3) -115./16.*ZETA3 +5./4.*ZETA2*ZETA3 +ZETA3*ZETA3/2. +83./4.*ZETA5)
	+CF*CF*CA*(74321./2304. -6593./864.*ZETA2 +376./135.*ZETA2*ZETA2 -2309./1260.*pow(ZETA2,3) -8653./432.*ZETA3 +53./9.*ZETA2*ZETA3 +37./12.*ZETA3*ZETA3 -689./72.*ZETA5)
	+CF*CA*CA*(-1505881./62208. +843./64.*ZETA2 +14611./8640.*ZETA2*ZETA2 +829./1260.*pow(ZETA2,3) +82385./5184.*ZETA3 -221./48.*ZETA2*ZETA3 -25./12.*ZETA3*ZETA3 -51./16.*ZETA5)
	+CF*CA*nf*(110651./15552. -7033./1296.*ZETA2 -1439./2160.*ZETA2*ZETA2 -94./81.*ZETA3 +13./12.*ZETA2*ZETA3 -ZETA5/8.)
	+CF*CF*nf*(-421./192. +329./216.*ZETA2 -223./540.*ZETA2*ZETA2 +869./216.*ZETA3 -14./9.*ZETA2*ZETA3 -19./18.*ZETA5)
	+CF*nf*nf*(-7081./15552. +151./324.*ZETA2 +2./27.*ZETA2*ZETA2 -79./324.*ZETA3)
	+CF*NFV*(CA*CA-4)/CA*(1./8. + 5./16.*ZETA2 -1./80.*ZETA2*ZETA2 +7./48.*ZETA3 -5./6.*ZETA5);
      d3c = pow(A1,3)*(160./3.*ZETA3*ZETA3-928./105.*pow(ZETA2,3))
	- pow(A1*ZETA2,2)*4./5.*d1c - 8./5.*ZETA2*ZETA2*A1*A2
	+ 4*ZETA3*A1*D2 + 32.*A1*A1*b0*(ZETA2*ZETA3-2.*ZETA5);
      d3c *= pow(M_PI,3);
      d3c += Cth3;
    }
    double d1 = d1c;
    double d2 = d2c;
    double d3 = d3c;
    double beta0 = b0*M_PI;
    double beta1 = b1*b0*M_PI*M_PI;
    //double beta2 = b2*b0*M_PI*M_PI*M_PI;
    //
    // write P_qq(x) = gA * D0(x) + gB * delta(1-x) + ...
    if(forceredo||!(last_nf==nf)) {
      gA0 = CF;
      gA1 = CF*(CA*(67./36.-ZETA2/2.) - 5./18.*nf);  // ok
      gB0 = CF*3./4.;
      gB1 = CF*CA/4.*(17./24.+11./3.*ZETA2-3.*ZETA3) - CF*nf/4.*(1./12.+2./3.*ZETA2) +CF*CF/4.*(3./8.-3.*ZETA2+6*ZETA3); //ok
      gB2 = CF*CF*CA/4.*(151./64.+ZETA2*ZETA3-205./24.*ZETA2-247./60.*ZETA2*ZETA2+211./12.*ZETA3+15./2.*ZETA5)
	+ CF*CA*nf/4.*(5./4.-167.*ZETA2/54.+ZETA2*ZETA2/20.+25./18.*ZETA3)
	- CF*CA*CA/4.*(1657./576.-281./27.*ZETA2+ZETA2*ZETA2/8.+97./9.*ZETA3-5./2.*ZETA5)
	- CF*nf*nf/4.*(17./144.-5./27.*ZETA2+ZETA3/9.)
	- CF*CF*nf/4.*(23./16.-5./12.*ZETA2-29./30.*ZETA2*ZETA2+17./6.*ZETA3)
	+ CF*CF*CF/4.*(29./32.-2*ZETA2*ZETA3+9./8.*ZETA2+18./5.*ZETA2*ZETA2+17./4.*ZETA3-15.*ZETA5);  // check this! eq 4.9 of hep-ph/0403192
    }
    //
    if(forceredo||!(last_nf==nf && last_M==M && last_mtop==Hmtop && last_EFT==mtEFT && last_muFrat==muF/M && last_muRrat==muR/M)) {
      double lf  = 2.*log(M/muF);
      double lrf = 2.*log(muR/muF);
      double lr = 2.*log(muR/M);
      //
      // C = sum as^k a_kj D_j(z)
      double a11 = 4*CF;
      double a23 = 8*CF*CF;
      double a22 = -4*CF*beta0;
      double a21 = 4*CF*d1 + CF*(CA*67./9.-(16*CF+2*CA)*ZETA2 - 10./9.*nf); // MB: check this! (should be correct)
      double a20 = CA*CF*(-101./27.+11./3.*ZETA2+7./2.*ZETA3) + CF*CF*16*ZETA3 + CF*nf*(14./27.-2./3.*ZETA2); // MB: check this! (should be correct)
      // factorization scale dependence (as=as(M))
      d3 += (4./3.*gB0*gB0*gB0 + 2*gB0*gB0*beta0 + 2./3.*gB0*beta0*beta0 - 4*gA0*gA0*gB0*ZETA2 -2*gA0*gA0*beta0*ZETA2 + 8./3.*gA0*gA0*gA0*ZETA3)*lf*lf*lf
	+ (4*gB1*gB0+2*gB1*beta0+gB0*beta1 - 4*gA0*gA1*ZETA2 + gA0*a11*(4*gB0+beta0)*ZETA3 + 2*a11*gA0*gA0*(ZETA2*ZETA2-3*ZETA4) + d1*(2*gB0*gB0+gB0*beta0-2*gA0*gA0*ZETA2))*lf*lf
	+ (2*gB2 +2*a11*gA1*ZETA3+2*gB1*d1 +2*gB0*d2+2*gA0*(6*ZETA5*a23-2*ZETA4*a22+ZETA3*a21-ZETA2*a20))*lf;
      d2 += (-2.*gA0*gA0*ZETA2 + gB0*(2*gB0+beta0))*lf*lf + (2*gB1+2*gB0*d1+2*a11*gA0*ZETA3)*lf;
      d1 += 2*gB0*lf;
      //
      // adding renormalization scale dependence
      d3 += (d1*beta1+d2*2*beta0)*lr + d1*beta0*beta0*lr*lr;
      d2 += d1*beta0*lr;
      //
      // restore normalization
      d1 /= M_PI;
      d2 /= M_PI*M_PI;
      d3 /= M_PI*M_PI*M_PI;
      //
      // construct full g0k
      //const double c1 = EulerGamma;
      double c2 = Power(EulerGamma,2) -ZETA2;
      double c3 = Power(EulerGamma,3) -3*EulerGamma*ZETA2 +2*ZETA3;
      double c4 = Power(EulerGamma,4) -6*EulerGamma*EulerGamma*ZETA2 +8*EulerGamma*ZETA3 +3*ZETA2*ZETA2 -6*ZETA4;
      double c5 = 24*ZETA5 +3*EulerGamma*ZETA2*ZETA2 +Power(EulerGamma,5) - 10*Power(EulerGamma,3)*ZETA2 +20*EulerGamma*EulerGamma*ZETA3 -20*ZETA3*ZETA2;
      double c6 = 144*ZETA5*EulerGamma +Power(EulerGamma,6) -15*Power(EulerGamma,4)*ZETA2 +40*Power(EulerGamma,3)*ZETA3 +9*EulerGamma*EulerGamma*ZETA2*ZETA2 -120*EulerGamma*ZETA3*ZETA2 +40*ZETA3*ZETA3 -45./7.*Power(ZETA2,3);
      //
      double S1 = 2*A1* (ZETA2+EulerGamma*EulerGamma - EulerGamma*lf);
      double S2 = 4*A1*b0/3.*c3 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c2
	- EulerGamma* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
			- 4.*A1*b0*EulerGamma*lr +2*A2*lf
			- A1*b0*(lr*lr - lrf*lrf) );
      double S3 = -c4* (4*A1*Power(b0,2))/3.
	+c3* (4*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma) - 4*A1*b0*log(M/muR)))/3.
	-c2* 2*(A3 + b0*(-D2 + 4*A2*EulerGamma + 
			 2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))) - 
		2*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*log(M/muR) + 
		4*A1*Power(b0,2)*Power(log(M/muR),2))
	-EulerGamma* (3*D3 - 12*A3*EulerGamma + 12*b0*D2*EulerGamma - 24*A2*b0*Power(EulerGamma,2) - 
		      12*A1*b0*b1*Power(EulerGamma,2) - 16*A1*Power(b0,2)*Power(EulerGamma,3) - 
		      24*A2*b0*ZETA2 - 12*A1*b0*b1*ZETA2 - 48*A1*Power(b0,2)*EulerGamma*ZETA2 - 
		      32*A1*Power(b0,2)*ZETA3 + 12*
		      (A3 + b0*(-D2 + 4*A2*EulerGamma + 
				2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))))*log(M/muR) - 
		      12*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*Power(log(M/muR),2) + 
		      16*A1*Power(b0,2)*Power(log(M/muR),3) - 12*A3*log(muF/muR) + 
		      12*b0*(2*A2 + A1*b1)*Power(log(muF/muR),2) - 
		      16*A1*Power(b0,2)*Power(log(muF/muR),3))/3.;
      // S1^2
      const double S1s = -4*A1*A1*(c4 -2*(2*EulerGamma-lf)*c3 +Power(2*EulerGamma-lf,2)*c2);
      // S1^3
      const double S1c = -8*A1*A1*A1*(c6 -3*(2*EulerGamma-lf)*c5 +3*Power(2*EulerGamma-lf,2)*c4 -Power(2*EulerGamma-lf,3)*c3);
      // S1*S2
      const double S1S2 = 2*A1*(4*A1*b0/3.*c5 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c4
				- c3* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
					- 4.*A1*b0*EulerGamma*lr +2*A2*lf
					- A1*b0*(lr*lr - lrf*lrf) ))
	-2*A1*(2*EulerGamma-lf)* (4*A1*b0/3.*c4 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c3
				  - c2* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
					  - 4.*A1*b0*EulerGamma*lr +2*A2*lf
					  - A1*b0*(lr*lr - lrf*lrf) ));
      //
      const double gD1 = S1;
      const double gD2 = S1s/2. + S2 + gD1*gD1;
      const double gD3 = S1c/6. + S1S2 + S3 + 2*gD1*gD2 - gD1*gD1*gD1;
      //
      g01 = d1 + gD1;
      g02 = d2 + d1*gD1 + gD2;
      g03 = d3 + d2*gD1 + d1*gD2 + gD3; 
      //
      //cout << "  g01 = " << g01 << endl;
      //cout << "  g02 = " << g02 << endl;
      //cout << "  g03 = " << g03 << endl;
      //
      g0sub1s = S1;
      g0sub2s = g01*S1 + S1s/2. + S2;
      g0sub3s = g01*(S1s/2.+S2) + g02*S1 + S1c/6. + S1S2 + S3;
      //
      g0bar1 = d1;
      g0bar2 = g02 - g01*S1 - S2 + S1*S1/2.;
      g0bar3 = g03 - g02*S1 - g01*(S2-S1*S1/2.) - (S1*S1*S1/6. - S1*S2 + S3);
    }
    //
    g0sub1 = 0;
    g0sub2 = 0;
    g0sub3 = 0;
    //
    if(!expBP && expDconst && prescr>0) { // in fact, this is a stupid option, so it should never happen
      g0sub1 = g0sub1s;
      g0sub2 = g0sub2s;
      g0sub3 = g0sub3s;
    }
    if(expBP && expDconst) {
      //cout << "g0bar1 = " << g0bar1 << endl;
      //cout << "g0bar2 = " << g0bar2 << endl;
      //cout << "g0bar3 = " << g0bar3 << endl;
      g01 = g0bar1;
      g02 = g0bar2;
      g03 = g0bar3;
    }
    /*
    // old version
    cout << "g01 = "<< g01 << endl;
    g01 = CF*(4.*ZETA2-4.+2.*EulerGamma*EulerGamma + (4.*EulerGamma-3.)*log(muF/M))/M_PI;
    cout << "g01 = "<< g01 << endl; // ok, with scale dep
    cout << "g02 = "<< g02 << endl;
    g02 = ( CF * ( 511./4. - 198.*ZETA2 - 60.*ZETA3 + 552.*ZETA2*ZETA2/5.
		    + 128.*EulerGamma*EulerGamma*(ZETA2-1.+EulerGamma*EulerGamma/4.) )
	    + CA * ( -1535./12. + 376.*ZETA2/3. + 604.*ZETA3/9. - 92.*ZETA2*ZETA2/5. + 1616.*EulerGamma/27.
			   - 56.*EulerGamma*ZETA3 + EulerGamma*EulerGamma* (536./9. - 16.*ZETA2 + 176.*EulerGamma/9.) )
	    + nf * ( 127./6. - 64.*ZETA2/3. + 8.*ZETA3/9. - EulerGamma*(224./27.+EulerGamma*(80.+32.*EulerGamma)/9.) )
	    + 4.*SQR(log(muF/M)) * ( CF * (32.*EulerGamma*EulerGamma - 48.*EulerGamma +18.)
					   + CA * (44.*EulerGamma/3.-11.)
					   + nf * (2.-8.*EulerGamma/3.))
	    - 2.*log(muF/M) * ( CF * (48.*ZETA3 + 72.*ZETA2 -93. -128.*EulerGamma*ZETA2 +128.*EulerGamma
						  + 48.*EulerGamma*EulerGamma -64.*EulerGamma*EulerGamma*EulerGamma)
				       + CA * (193./3. -24.*ZETA3 -88.*ZETA2/3. +16.*EulerGamma*ZETA2 -536*EulerGamma/9.
						-88.*EulerGamma*EulerGamma/3.)
				       + nf * (16.*ZETA2/3. -34./3. +80.*EulerGamma/9. +16.*EulerGamma*EulerGamma/3.))
	    )*CF/16./M_PI/M_PI  - b0*g01*2.*log(muF/muR);
    cout << "g02 = "<< g02 << endl<< endl;  // ok, with scale dep
    g03 = 0; // unknown? or known maybe: 1404.0366
    //*/
    return;
  }


  void Resummation::InitConstHiggsCommons(bool pseudoscalar) {
  using namespace QCD;
  if(!pseudoscalar) { // SM Higgs
    // EFT
    double PLd1 = 2.*CA * ZETA2;
    double PLd2 = ( 837./16. + 67./2.*ZETA2 - 9./20.*ZETA2*ZETA2 - 165./4.*ZETA3
		    + nf * ( 5./6.*ZETA3 - 5./3.*ZETA2 - 247./36. ) );
    double PLd3 = 1740.02 - 131.928*nf + 1.75719*nf*nf; // new coefficient
    double lHt = 2.*log(M/Hmtop);
    W1 = 11./2.;
    W2 = 121./16. + 2777./144. + 19./8.*lHt + (-67./48. + 2./3.*lHt)*nf;
    // This is with MSbar top mass! (1403.4616)
    W3 = (-1792967./20736. +  897943./4608.*ZETA3 + 5347./288.*lHt + 209./32.*lHt*lHt
	  + nf * (493./10368. + 209./54.*lHt + 23./16.*lHt*lHt - 110779./6912.*ZETA3)
	  + nf*nf * (-6865./15552. + 77./864.*lHt - lHt*lHt/9. ) );
    // conversion to pole top mass
    W3 += 2*(19./8. + 2./3.*nf)*(4./3. + lHt);
    //
    d1c = PLd1 + W1;
    d2c = PLd2 + PLd1*W1 + W2;
    d3c = PLd3 + PLd2*W1 + PLd1*W2 + W3;
    //
    if(!mtEFT) {
      W1 = higgs_nlo_G2l(M, Hmtop, Hmbot, Hmcha);
      if(M < 2*Hmtop) W2 = higgs_nnlo_delta_term(M, Hmtop, nf) - PLd2 - PLd1*W1;
      W3 = d3c - PLd3 - PLd2*W1 - PLd1*W2; // if commented uses the EFT value of W3 (consistent with ggHiggs now)
      //
      // recompute d1c, d2c, d3c accordingly
      d1c = PLd1 + W1;
      d2c = PLd2 + PLd1*W1 + W2;
      d3c = PLd3 + PLd2*W1 + PLd1*W2 + W3;
    }
  } else { // pseudo scalar
    // working in the large-mt EFT
    double Ltop = 2*log(M/Hmtop); 
    double CJ2 = 0.;
    d1c = CA*(2.+2.*ZETA2);
    d2c = CA*CA*(494./3+1112./9*ZETA2-4./5.*pow(ZETA2,2)-220./3*ZETA3)
      +   CA*nf*(-82./3-80./9*ZETA2-8./3*ZETA3)
      +   CF*nf*(-160./3+12*Ltop+16.*ZETA3);
    d2c /= 16.;
    d3c = nf*CJ2*(-4.)
      +CF*nf*nf*(1498./9-40./9*ZETA2-32./45*pow(ZETA2,2)-224./3*ZETA3)
      +CF*CF*nf*(457./3+208.*ZETA3-320.*ZETA5)
      +CA*CA*nf*(-113366./81-10888./81*ZETA2+21032./135*pow(ZETA2,2)+8840./27*ZETA3-2000./3*ZETA2*ZETA3+6952./9*ZETA5)
      +CA*CA*CA*(114568./27+137756./81*ZETA2-61892./135*pow(ZETA2,2)-64096./105*pow(ZETA2,3)-3932*ZETA3+7832./3*ZETA2*ZETA3+13216./3*pow(ZETA3,2)-30316./9*ZETA5)
      +CA*nf*nf*(6914./81-1696./81*ZETA2-608./45*pow(ZETA2,2)+688./27*ZETA3)
      +CA*CF*nf*(-1797.-4160./9*ZETA2+176./45*pow(ZETA2,2)+1856./3*ZETA3+192.*ZETA2*ZETA3+160.*ZETA5+96.*Ltop*(1+ZETA2));
    d3c /= 64.;
  }
  return;
}


  void Resummation::InitConstHiggs(bool pseudoscalar) {
    using namespace QCD;
    //
    double beta0 = b0*M_PI;
    double beta1 = b1*b0*M_PI*M_PI;
    double beta2 = b2*b0*M_PI*M_PI*M_PI;
    //
    if(forceredo||!(last_nf==nf && last_M==M && last_mtop==Hmtop && last_EFT==mtEFT)) {
      InitConstHiggsCommons(pseudoscalar);
    }
    //
    double d1 = d1c;
    double d2 = d2c;
    double d3 = d3c;
    //
    // write P_gg(x) = gA * D0(x) + gB * delta(1-x) + ...
    if(forceredo||!(last_nf==nf)) {
      gA0 = CA;
      gA1 = CA*CA*(67./36.-ZETA2/2.) - 5./18.*CA*nf;
      //gA2 = CA*CA*CA*(245./24.-67./9.*ZETA2+11./6.*ZETA3+11./5.*ZETA2*ZETA2)/4.
      //  + CA*CA*nf*(-209./108.+10./9.*ZETA2-7./3.*ZETA3)/4.
      //  + CA*CF*nf*(-55./24.+2.*ZETA3)/4.
      //  - CA*nf*nf/4./27.;
      gB0 = beta0;
      gB1 = CA*CA*(2./3.+3./4.*ZETA3) - (CA/6.+CF/8.)*nf;
      gB2 = CA*CA*CA/4.*(79./32.-ZETA2*ZETA3+ZETA2/6.+11./24.*ZETA2*ZETA2+67./6.*ZETA3-5.*ZETA5)
	- CA*CA*nf/4.*(233./288.+ZETA2/6.+ZETA2*ZETA2/12.+5./3.*ZETA3)
	- CA*CF*nf/4.*241./288.
	+ CF*CF*nf/4./16.
	+ nf*nf/4.*(CA*29./288.+CF*11./144.);
      //cout << gB2*4*4*4 << "   " << 4425.894 - nf*528.723 + nf*nf*6.4630 << endl;
      //gB2 = (4425.894 - nf*528.723 + nf*nf*6.4630)/4./4./4.;
    }
    //
    if(forceredo||!(last_nf==nf && last_M==M && last_mtop==Hmtop && last_EFT==mtEFT && last_muFrat==muF/M && last_muRrat==muR/M)) {
      double lf  = 2.*log(M/muF);
      double lrf = 2.*log(muR/muF);
      double lr = 2.*log(muR/M);
      //cout << lf << "  " << -lr << endl;
      //
      // C = sum as^k a_kj D_j(z)
      double a11 = 4*CA;
      double a23 = 8*CA*CA;
      double a22 = -4*CA*beta0;
      double a21 = 4*CA*d1 + CA*CA*(67./9.-18.*ZETA2) - 10./9.*CA*nf;
      double a20 = CA*CA*(-101./27.+11./3.*ZETA2+39./2.*ZETA3) + CA*nf*(14./27.-2./3.*ZETA2);
      // factorization scale dependence (as=as(M))
      d3 += (4./3.*gB0*gB0*gB0 + 2*gB0*gB0*beta0 + 2./3.*gB0*beta0*beta0 - 4*gA0*gA0*gB0*ZETA2 -2*gA0*gA0*beta0*ZETA2 + 8./3.*gA0*gA0*gA0*ZETA3)*lf*lf*lf
	+ (4*gB1*gB0+2*gB1*beta0+gB0*beta1 - 4*gA0*gA1*ZETA2 + gA0*a11*(4*gB0+beta0)*ZETA3 + 2*a11*gA0*gA0*(ZETA2*ZETA2-3*ZETA4) + d1*(2*gB0*gB0+gB0*beta0-2*gA0*gA0*ZETA2))*lf*lf
	+ (2*gB2 +2*a11*gA1*ZETA3+2*gB1*d1 +2*gB0*d2+2*gA0*(6*ZETA5*a23-2*ZETA4*a22+ZETA3*a21-ZETA2*a20))*lf;
      d2 += (-2.*gA0*gA0*ZETA2 + gB0*(2*gB0+beta0))*lf*lf + (2*gB1+2*gB0*d1+2*a11*gA0*ZETA3)*lf;
      d1 += 2*gB0*lf;
      /*
      // correct as well, but makes explicit use of gB0=beta0, gA0=CA and a11=4CA
      d3 += (4*beta0*beta0*beta0 - 6*CA*CA*beta0*ZETA2 + 8./3.*CA*CA*CA*ZETA3)*lf*lf*lf
	+ (beta0*(6*gB1+beta1) - 4*CA*gA1*ZETA2 + 20*CA*CA*beta0*ZETA3 + 8*CA*CA*CA*(ZETA2*ZETA2-3*ZETA4) + d1*(3*beta0*beta0-2*CA*CA*ZETA2))*lf*lf
	+ (2*gB2+8*CA*gA1*ZETA3+2*gB1*d1+2*beta0*d2+2*CA*(6*ZETA5*a23-2*ZETA4*a22+ZETA3*a21-ZETA2*a20))*lf;
      d2 += (-2.*CA*CA*ZETA2 + 3*beta0*beta0)*lf*lf + (2*gB1+2*beta0*d1+8*CA*CA*ZETA3)*lf;
      d1 += 2*beta0*lf;
      */
      //
      // adding renormalization scale dependence
      d3 += (2*beta2+3*beta1*d1+4*beta0*d2)*lr + (7*beta0*beta1+6*beta0*beta0*d1)*lr*lr + 4*Power(beta0*lr,3);
      d2 += (2*beta1+3*beta0*d1)*lr + 3*beta0*beta0*lr*lr;
      d1 += 2*beta0*lr;
      //
      //cout << "d1-d1(central) = " << (d1-d1c)/M_PI << endl;
      //cout << "d2-d2(central) = " << (d2-d2c)/M_PI/M_PI << endl;
      //cout << "d3-d3(central) = " << (d3-d3c)/M_PI/M_PI/M_PI << endl;
      //
      // restore normalization
      d1 /= M_PI;
      d2 /= M_PI*M_PI;
      d3 /= M_PI*M_PI*M_PI;
      //cout << "d1 = " << d1 << endl;
      //cout << "d2 = " << d2 << endl;
      //cout << "d3 = " << d3 << endl;
      //
      // construct full g0k
      //const double c1 = EulerGamma;
      double c2 = Power(EulerGamma,2) -ZETA2;
      double c3 = Power(EulerGamma,3) -3*EulerGamma*ZETA2 +2*ZETA3;
      double c4 = Power(EulerGamma,4) -6*EulerGamma*EulerGamma*ZETA2 +8*EulerGamma*ZETA3 +3*ZETA2*ZETA2 -6*ZETA4;
      double c5 = 24*ZETA5 +3*EulerGamma*ZETA2*ZETA2 +Power(EulerGamma,5) - 10*Power(EulerGamma,3)*ZETA2 +20*EulerGamma*EulerGamma*ZETA3 -20*ZETA3*ZETA2;
      double c6 = 144*ZETA5*EulerGamma +Power(EulerGamma,6) -15*Power(EulerGamma,4)*ZETA2 +40*Power(EulerGamma,3)*ZETA3 +9*EulerGamma*EulerGamma*ZETA2*ZETA2 -120*EulerGamma*ZETA3*ZETA2 +40*ZETA3*ZETA3 -45./7.*Power(ZETA2,3);
      //
      double S1 = 2*A1* (ZETA2+EulerGamma*EulerGamma - EulerGamma*lf);
      double S2 = 4*A1*b0/3.*c3 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c2
	- EulerGamma* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
			- 4.*A1*b0*EulerGamma*lr +2*A2*lf
			- A1*b0*(lr*lr - lrf*lrf) );
      double S3 = -c4* (4*A1*Power(b0,2))/3.
	+c3* (4*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma) - 4*A1*b0*log(M/muR)))/3.
	-c2* 2*(A3 + b0*(-D2 + 4*A2*EulerGamma + 
			 2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))) - 
		2*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*log(M/muR) + 
		4*A1*Power(b0,2)*Power(log(M/muR),2))
	-EulerGamma* (3*D3 - 12*A3*EulerGamma + 12*b0*D2*EulerGamma - 24*A2*b0*Power(EulerGamma,2) - 
		      12*A1*b0*b1*Power(EulerGamma,2) - 16*A1*Power(b0,2)*Power(EulerGamma,3) - 
		      24*A2*b0*ZETA2 - 12*A1*b0*b1*ZETA2 - 48*A1*Power(b0,2)*EulerGamma*ZETA2 - 
		      32*A1*Power(b0,2)*ZETA3 + 12*
		      (A3 + b0*(-D2 + 4*A2*EulerGamma + 
				2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))))*log(M/muR) - 
		      12*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*Power(log(M/muR),2) + 
		      16*A1*Power(b0,2)*Power(log(M/muR),3) - 12*A3*log(muF/muR) + 
		      12*b0*(2*A2 + A1*b1)*Power(log(muF/muR),2) - 
		      16*A1*Power(b0,2)*Power(log(muF/muR),3))/3.;
      // S1^2
      const double S1s = -4*A1*A1*(c4 -2*(2*EulerGamma-lf)*c3 +Power(2*EulerGamma-lf,2)*c2);
      // S1^3
      const double S1c = -8*A1*A1*A1*(c6 -3*(2*EulerGamma-lf)*c5 +3*Power(2*EulerGamma-lf,2)*c4 -Power(2*EulerGamma-lf,3)*c3);
      // S1*S2
      const double S1S2 = 2*A1*(4*A1*b0/3.*c5 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c4
				- c3* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
					- 4.*A1*b0*EulerGamma*lr +2*A2*lf
					- A1*b0*(lr*lr - lrf*lrf) ))
	-2*A1*(2*EulerGamma-lf)* (4*A1*b0/3.*c4 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c3
				  - c2* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
					  - 4.*A1*b0*EulerGamma*lr +2*A2*lf
					  - A1*b0*(lr*lr - lrf*lrf) ));
      //
      const double gD1 = S1;
      const double gD2 = S1s/2. + S2 + gD1*gD1;
      const double gD3 = S1c/6. + S1S2 + S3 + 2*gD1*gD2 - gD1*gD1*gD1;
      //
      g01 = d1 + gD1;
      g02 = d2 + d1*gD1 + gD2;
      g03 = d3 + d2*gD1 + d1*gD2 + gD3; 
      //
      g0sub1s = S1;
      g0sub2s = g01*S1 + S1s/2. + S2;
      g0sub3s = g01*(S1s/2.+S2) + g02*S1 + S1c/6. + S1S2 + S3;
      //
      /*/
      c2 += ZETA2;
      c3 += -(1.5)*ZETA3;
      c4 += 27./5.*ZETA2*ZETA2;
      c5 += -(22.5)*ZETA5-(15.0)*ZETA2*ZETA3;
      c6 += 549./7.*ZETA2*ZETA2*ZETA2+(22.5)*ZETA3*ZETA3;
      //
      S1 = 2*A1* (ZETA2+EulerGamma*EulerGamma - EulerGamma*lf);
      S2 = 4*A1*b0/3.*c3 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c2
	- EulerGamma* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
			- 4.*A1*b0*EulerGamma*lr +2*A2*lf
			- A1*b0*(lr*lr - lrf*lrf) );
      S3 = -c4* (4*A1*Power(b0,2))/3.
	+c3* (4*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma) - 4*A1*b0*log(M/muR)))/3.
	-c2* 2*(A3 + b0*(-D2 + 4*A2*EulerGamma + 
			 2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))) - 
		2*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*log(M/muR) + 
		4*A1*Power(b0,2)*Power(log(M/muR),2))
	-EulerGamma* (3*D3 - 12*A3*EulerGamma + 12*b0*D2*EulerGamma - 24*A2*b0*Power(EulerGamma,2) - 
		      12*A1*b0*b1*Power(EulerGamma,2) - 16*A1*Power(b0,2)*Power(EulerGamma,3) - 
		      24*A2*b0*ZETA2 - 12*A1*b0*b1*ZETA2 - 48*A1*Power(b0,2)*EulerGamma*ZETA2 - 
		      32*A1*Power(b0,2)*ZETA3 + 12*
		      (A3 + b0*(-D2 + 4*A2*EulerGamma + 
				2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))))*log(M/muR) - 
		      12*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*Power(log(M/muR),2) + 
		      16*A1*Power(b0,2)*Power(log(M/muR),3) - 12*A3*log(muF/muR) + 
		      12*b0*(2*A2 + A1*b1)*Power(log(muF/muR),2) - 
		      16*A1*Power(b0,2)*Power(log(muF/muR),3))/3.;
      //*/
      g0bar1 = d1;
      g0bar2 = g02 - g01*S1 - S2 + S1*S1/2.;
      g0bar3 = g03 - g02*S1 - g01*(S2-S1*S1/2.) - (S1*S1*S1/6. - S1*S2 + S3);
      //cout << "g0bar1 = " << g0bar1 << endl;
      //cout << "g0bar2 = " << g0bar2 << endl;
      //cout << "g0bar3 = " << g0bar3 << endl;
    }
    //
    //cout << "gD1 = " << gD1 << endl;
    //cout << "gD2 = " << gD2 << endl;
    //cout << "gD3 = " << gD3 << endl;
    //cout << "g01 = " << g01 << endl;
    //cout << "g02 = " << g02 << endl;
    //cout << "g03 = " << g03 << endl;
    //
    //
    g0sub1 = 0;
    g0sub2 = 0;
    g0sub3 = 0;
    //
    if(!expBP && expDconst && prescr>0) {
      cout << endl << "BAD PRESCRIPTION if you ended up here!!" << endl << endl;
      g0sub1 = g0sub1s;
      g0sub2 = g0sub2s;
      g0sub3 = g0sub3s;
    }

    if(expBP && expDconst) {
      //cout << "g0bar1 = " << g0bar1 << endl;
      //cout << "g0bar2 = " << g0bar2 << endl;
      //cout << "g0bar3 = " << g0bar3 << endl;
      g01 = g0bar1;
      g02 = g0bar2;
      g03 = g0bar3;
    }
    //
    return;
  }


  void Resummation::InitConstDIS() {
    using namespace QCD;
    //
    // coefficients of the delta(1-z) at central scales (from hep-ph/0504242)
    if(forceredo||!(last_nf==nf && last_M==M && last_mtop==Hmtop && last_EFT==mtEFT)) {
      d1c = -CF*(ZETA2+9./4.);
      d2c = (-338.513 + nf*46.8531)/16.;
      d3c = (-9170.38+nf*2574.687-nf*nf*103.2366)/64.;
      double fl11ns = 0; // = 3 <e>, see tab 2 of hep-ph/0504242
      d3c += -fl11ns*nf*11.888/64.;
    }
    double d1 = d1c;
    double d2 = d2c;
    double d3 = d3c;
    double beta0 = b0*M_PI;
    double beta1 = b1*b0*M_PI*M_PI;
    //double beta2 = b2*b0*M_PI*M_PI*M_PI;
    //
    // write P_qq(x) = gA * D0(x) + gB * delta(1-x) + ...
    if(forceredo||!(last_nf==nf)) {
      gA0 = CF;
      gA1 = CF*(CA*(67./36.-ZETA2/2.) - 5./18.*nf);
      gB0 = CF*3./4.;
      gB1 = CF*CA/4.*(17./24.+11./3.*ZETA2-3.*ZETA3) - CF*nf/4.*(1./12.+2./3.*ZETA2) +CF*CF/4.*(3./8.-3.*ZETA2+6*ZETA3);
      gB2 = CF*CF*CA/4.*(151./64.+ZETA2*ZETA3-205./24.*ZETA2-247./60.*ZETA2*ZETA2+211./12.*ZETA3+15./2.*ZETA5)
	+ CF*CA*nf/4.*(5./4.-167.*ZETA2/54.+ZETA2*ZETA2/20.+25./18.*ZETA3)
	- CF*CA*CA/4.*(1657./576.-281./27.*ZETA2+ZETA2*ZETA2/8.+97./9.*ZETA3-5./2.*ZETA5)
	- CF*nf*nf/4.*(17./144.-5./27.*ZETA2+ZETA3/9.)
	- CF*CF*nf/4.*(23./16.-5./12.*ZETA2-29./30.*ZETA2*ZETA2+17./6.*ZETA3)
	+ CF*CF*CF/4.*(29./32.-2*ZETA2*ZETA3+9./8.*ZETA2+18./5.*ZETA2*ZETA2+17./4.*ZETA3-15.*ZETA5);
    }
    //
    if(forceredo||!(last_nf==nf && last_M==M && last_mtop==Hmtop && last_EFT==mtEFT && last_muFrat==muF/M && last_muRrat==muR/M)) {
      double lf  = 2.*log(M/muF);
      double lrf = 2.*log(muR/muF);
      double lr = 2.*log(muR/M);
      //
      // C = sum as^k a_kj D_j(z)   (from hep-ph/0504242)
      double a11 = CF;
      double a10 = -3./4.*CF;
      double a23 = 8./9.;
      double a22 = -23./6. + nf/9.;
      double a21 = -31.1052/16. -nf*29./54.;
      double a20 = 188.641/16. + nf*6.34888/16.;
      // factorization scale dependence (as=as(M))
      //MB these should be now correct
      d3 += (1./6.*gB0*gB0*gB0 + gB0*gB0*beta0/2. + 1./3.*gB0*beta0*beta0 - gA0*gA0*gB0*ZETA2/2. -gA0*gA0*beta0*ZETA2/2. + 1./3.*gA0*gA0*gA0*ZETA3)*lf*lf*lf
	+ (gB1*gB0+gB1*beta0+gB0/2.*beta1 - gA0*gA1*ZETA2 + gA0*a11*(gB0+beta0/2.)*ZETA3 -gA0*a10*(gB0+beta0/2.)*ZETA2 + a11*gA0*gA0*(ZETA2*ZETA2-3*ZETA4)/2. +a10*gA0*gA0*ZETA3 + d1*(gB0*gB0+gB0*beta0-gA0*gA0*ZETA2)/2.)*lf*lf
	+ (gB2 +a11*gA1*ZETA3 -a10*gA1*ZETA2 +gB1*d1 +gB0*d2 +gA0*(6*ZETA5*a23-2*ZETA4*a22+ZETA3*a21-ZETA2*a20))*lf;
      d2 += (-gA0*gA0*ZETA2 + gB0*(gB0+beta0))/2.*lf*lf + (gB1+gB0*d1+a11*gA0*ZETA3-a10*gA0*ZETA2)*lf;
      d1 += gB0*lf;
      //
      // adding renormalization scale dependence
      d3 += (d1*beta1+d2*2*beta0)*lr + d1*beta0*beta0*lr*lr;
      d2 += d1*beta0*lr;
      //
      // restore normalization
      d1 /= M_PI;
      d2 /= M_PI*M_PI;
      d3 /= M_PI*M_PI*M_PI;
      //cout << d1 << endl;  // ok
      //cout << d2 << endl;  // ok
      //cout << d3 << endl;  // ok
      //
      // construct full g0k  (see g4.nb)
      // these are the derivatives Delta^{(k)}(1) with Delta(N)=1/Gamma(N) (see factorization_scale_dependence.cc)
      double c1 = EulerGamma;
      double c2 = GD(2,1);//Power(EulerGamma,2) -ZETA2;
      double c3 = GD(3,1);//Power(EulerGamma,3) -3*EulerGamma*ZETA2 +2*ZETA3;
      double c4 = GD(4,1);//Power(EulerGamma,4) -6*EulerGamma*EulerGamma*ZETA2 +8*EulerGamma*ZETA3 +3*ZETA2*ZETA2 -6*ZETA4;
      double c5 = GD(5,1);//24*ZETA5 +3*EulerGamma*ZETA2*ZETA2 +Power(EulerGamma,5) - 10*Power(EulerGamma,3)*ZETA2 +20*EulerGamma*EulerGamma*ZETA3 -20*ZETA3*ZETA2;
      double c6 = GD(6,1);//144*ZETA5*EulerGamma +Power(EulerGamma,6) -15*Power(EulerGamma,4)*ZETA2 +40*Power(EulerGamma,3)*ZETA3 +9*EulerGamma*EulerGamma*ZETA2*ZETA2 -120*EulerGamma*ZETA3*ZETA2 +40*ZETA3*ZETA3 -45./7.*Power(ZETA2,3);
      //
      // here we use the expansion of S, with the replacement L^k -> -Delta^{(k)}(1)=ck
      double S12 = A1/2.;
      double S11 = (D1-4*A1*EG +2*A1*lf)/2.;
      double S1 = -c2*S12 - c1*S11;
      double S23 = -A1*b0/6.;
      double S22 = A1*b0*(EG+lr/2.) + A2/2. - b0*D1/4.;
      double S21 = (D2/2. - 2.*A2*EG + b0*D1*EG - 2.*A1*b0*(EG*EG+Z2)
		    +(b0*D1 - 4.*A1*b0*EG)*lr/2. +A2*lf - A1*b0*(lr*lr - lrf*lrf)/2.);
      double S2 = -c3*S23 -c2*S22 -c1*S21;
      // MB S3 to be done
      double S3 = -c4*A1*b0*b0/12.
	+c3* b0*(2*A2 + A1*b1 - b0*D1 + 4*A1*b0*EG - 4*A1*b0*log(M/muR))/6.
	-c2* (2*A3 + b0*(-(b1*D1) - 2*D2 + 8*A2*EG + 4*A1*b1*EG - 4*b0*D1*EG + 8*A1*b0*Power(EG,2) + 
			 8*A1*b0*Z2) - 4*b0*(2*A2 + A1*b1 - b0*D1 + 4*A1*b0*EG)*log(M/muR) + 
	      8*A1*Power(b0,2)*Power(log(M/muR),2))/4.
	-c1* (3*D3 - 12*A3*EG + 6*b0*b1*D1*EG + 12*b0*D2*EG - 24*A2*b0*Power(EG,2) - 12*A1*b0*b1*Power(EG,2) + 
	      12*Power(b0,2)*D1*Power(EG,2) - 16*A1*Power(b0,2)*Power(EG,3) - 24*A2*b0*Z2 - 12*A1*b0*b1*Z2 + 
	      12*Power(b0,2)*D1*Z2 - 48*A1*Power(b0,2)*EG*Z2 - 32*A1*Power(b0,2)*Z3 + 
	      6*(2*A3 + b0*(-(b1*D1) - 2*D2 + 8*A2*EG + 4*A1*b1*EG - 4*b0*D1*EG + 8*A1*b0*Power(EG,2) + 
			    8*A1*b0*Z2))*log(M/muR) - 12*b0*(2*A2 + A1*b1 - b0*D1 + 4*A1*b0*EG)*Power(log(M/muR),2) + 
	      16*A1*Power(b0,2)*Power(log(M/muR),3) - 12*A3*log(muF/muR) + 
	      12*b0*(2*A2 + A1*b1)*Power(log(muF/muR),2) - 16*A1*Power(b0,2)*Power(log(muF/muR),3))/6.;
      // S1^2
      const double S1s = -c4*S12*S12 -2*c3*S12*S11 -c2*S11*S11;
      // S1^3
      const double S1c = -c6*S12*S12*S12 -3*c5*S11*S12*S12 -3*c4*S11*S11*S12 -c3*S11*S11*S11;
      // S1*S2
      const double S1S2 = -S12*(S23*c5+S22*c4+S21*c3) - S11*(S23*c4+S22*c3+S21*c2);
      //
      // g0constComputation = true;
      // LL_order = NNNLL;
      // cout << S1 << "   " << asS1(10,NULL)/as << endl;
      // cout << S2 << "   " << as2S2(10,NULL)/as/as << endl;
      // cout << S3 << "   " << as3S3(10,NULL)/as/as/as << endl;
      // S3 = real(as3S3(10,NULL)/as/as/as);
      //
      const double gD1 = S1;
      //const double gD2 = S1s/2. + S2 + gD1*gD1;
      //const double gD3 = S1c/6. + S1S2 + S3 + 2*gD1*gD2 - gD1*gD1*gD1;
      const double gD2 = S1s/2. + S2;
      const double gD3 = S1c/6. + S1S2 + S3;
      //
      g01 = d1 + gD1;
      //g02 = d2 + d1*gD1 + gD2;
      //g03 = d3 + d2*gD1 + d1*gD2 + gD3;
      g02 = d2 + g01*gD1 + gD2;
      g03 = d3 + g02*gD1 + g01*gD2 + gD3;
      //
      //cout << "  g01 = " << g01 << endl;
      //cout << "  g02 = " << g02 << endl;
      //cout << "  g03 = " << g03 << endl;
      //
      g0sub1s = S1;
      g0sub2s = g01*S1 + S1s/2. + S2;
      g0sub3s = g01*(S1s/2.+S2) + g02*S1 + S1c/6. + S1S2 + S3;
      //
      g0bar1 = d1;
      g0bar2 = g02 - g01*S1 - S2 + S1*S1/2.;
      g0bar3 = g03 - g02*S1 - g01*(S2-S1*S1/2.) - (S1*S1*S1/6. - S1*S2 + S3);
    }
    //
    g0sub1 = 0;
    g0sub2 = 0;
    g0sub3 = 0;
    //
    if(!expBP && expDconst && prescr>0) { // in fact, this is a stupid option, so it should never happen
      g0sub1 = g0sub1s;
      g0sub2 = g0sub2s;
      g0sub3 = g0sub3s;
    }
    if(expBP && expDconst) {
      //cout << "g0bar1 = " << g0bar1 << endl;
      //cout << "g0bar2 = " << g0bar2 << endl;
      //cout << "g0bar3 = " << g0bar3 << endl;
      g01 = g0bar1;
      g02 = g0bar2;
      g03 = g0bar3;
    }
    /*
    // old version
    cout << "g01 = " << g01 << endl;
    g01 = CF*(-9.-2*ZETA2+2*EulerGamma*EulerGamma+3*EulerGamma)/4./M_PI;
    cout << "g01 = " << g01 << endl;
    cout << "g02 = " << g02 << endl;
    g02 = (CF*CF*(331./8-51./2*EulerGamma-27./2*EulerGamma*EulerGamma+6*EulerGamma*EulerGamma*EulerGamma
		  +2*EulerGamma*EulerGamma*EulerGamma*EulerGamma+111./2*ZETA2
		  -18.*EulerGamma*ZETA2-4.*EulerGamma*EulerGamma*ZETA2-66.*ZETA3+24.*EulerGamma*ZETA3+4./5*ZETA2*ZETA2)
	   +CA*CF*(-5465./72+3155./54*EulerGamma+367./18*EulerGamma*EulerGamma+22./9*EulerGamma*EulerGamma*EulerGamma
		   -1139./18*ZETA2-22./3*EulerGamma*ZETA2-4.*EulerGamma*EulerGamma*ZETA2+464./9*ZETA3-40*EulerGamma*ZETA3+51./5*ZETA2*ZETA2)
	   +CF*nf*(457./36-247./27*EulerGamma-29./9*EulerGamma*EulerGamma-4./9*EulerGamma*EulerGamma*EulerGamma
		   +85./9*ZETA2+4./3.*EulerGamma*ZETA2+4./9*ZETA3)
	   )/4./M_PI/4./M_PI;
    cout << "g02 = " << g02 << endl;
    cout << "g03 = " << g03 << endl;
    //cout << nf << "   "<< g02*M_PI*M_PI << endl;
    //
    // this is d_{abc} d_{abc} / N_c * fl_{11}
    double fl11dabcdabcnc = 0;
    g03 = CF*nf*nf*(-19.58230452674897 + (8714*EulerGamma)/729. + (470*pow(EulerGamma,2))/81. + 
		    (116*pow(EulerGamma,3))/81. + (4*pow(EulerGamma,4))/27. - (2110*ZETA2)/81. - 
		    (116*EulerGamma*ZETA2)/27. - (8*pow(EulerGamma,2)*ZETA2)/9. - (292*pow(ZETA2,2))/135. + 
		    (80*ZETA3)/81. + (64*EulerGamma*ZETA3)/27.)
      + CF*CF*nf*(-9.472222222222221 + (2003*EulerGamma)/108. + (83*pow(EulerGamma,2))/18. - 
		  (683*pow(EulerGamma,3))/27. - (70*pow(EulerGamma,4))/9. - (8*pow(EulerGamma,5))/9. - 
		  (10733*ZETA2)/54. + (2177*EulerGamma*ZETA2)/27. + (112*pow(EulerGamma,2)*ZETA2)/3. + 
		  (32*pow(EulerGamma,3)*ZETA2)/9. - (10802*pow(ZETA2,2))/135. - 
		  (8*EulerGamma*pow(ZETA2,2))/3. + (10766*ZETA3)/27. - (20*EulerGamma*ZETA3)/9. + 
		  (8*pow(EulerGamma,2)*ZETA3)/9. - (40*ZETA2*ZETA3)/3. - (784*ZETA5)/9.)
      + CA*CF*nf*(293.99794238683126 - (160906*EulerGamma)/729. - (7531*pow(EulerGamma,2))/81. - 
		  (1552*pow(EulerGamma,3))/81. - (44*pow(EulerGamma,4))/27. + (33331*ZETA2)/81. + 
		  (5264*EulerGamma*ZETA2)/81. + (56*pow(EulerGamma,2)*ZETA2)/3. + (16*pow(EulerGamma,3)*ZETA2)/9. + 
		  (164*pow(ZETA2,2))/135. - (128*EulerGamma*pow(ZETA2,2))/15. - (21418*ZETA3)/81. + 
		  (1976*EulerGamma*ZETA3)/27. + 8*pow(EulerGamma,2)*ZETA3 - (64*ZETA2*ZETA3)/9. + (8*ZETA5)/3.)
      + CF*CF*CF*(-302.2916666666667 + (1001*EulerGamma)/8. + (187*pow(EulerGamma,2))/4. - 
		  (93*pow(EulerGamma,3))/2. - 9*pow(EulerGamma,4) + 6*pow(EulerGamma,5) + (4*pow(EulerGamma,6))/3. - 
		  (6197*ZETA2)/12. + (579*EulerGamma*ZETA2)/2. + 66*pow(EulerGamma,2)*ZETA2 - 
		  36*pow(EulerGamma,3)*ZETA2 - 4*pow(EulerGamma,4)*ZETA2 - (1791*pow(ZETA2,2))/5. + 
		  84*EulerGamma*pow(ZETA2,2) + (8*pow(EulerGamma,2)*pow(ZETA2,2))/5. + 
		  (8144*pow(ZETA2,3))/315. - 411*ZETA3 - 346*EulerGamma*ZETA3 - 60*pow(EulerGamma,2)*ZETA3 + 
		  48*pow(EulerGamma,3)*ZETA3 + 556*ZETA2*ZETA3 - 80*EulerGamma*ZETA2*ZETA3 - 
		  (176*pow(ZETA3,2))/3. + 1384*ZETA5 - 240*EulerGamma*ZETA5)
      + CA*CF*CF*(763.4166666666666 - (16981*EulerGamma)/24. - (5563*pow(EulerGamma,2))/36. + 
		  (8425*pow(EulerGamma,3))/54. + (433*pow(EulerGamma,4))/9. + (44*pow(EulerGamma,5))/9. + 
		  (191545*ZETA2)/108. - (28495*EulerGamma*ZETA2)/54. - (592*pow(EulerGamma,2)*ZETA2)/3. - 
		  (284*pow(EulerGamma,3)*ZETA2)/9. - 8*pow(EulerGamma,4)*ZETA2 + (11419*pow(ZETA2,2))/27. + 
		  (299*EulerGamma*pow(ZETA2,2))/3. + (142*pow(EulerGamma,2)*pow(ZETA2,2))/5. - 
		  (23098*pow(ZETA2,3))/315. - (49346*ZETA3)/27. + 752*EulerGamma*ZETA3 + 
		  (640*pow(EulerGamma,2)*ZETA3)/9. - 80*pow(EulerGamma,3)*ZETA3 - 828*ZETA2*ZETA3 + 
		  96*EulerGamma*ZETA2*ZETA3 + (536*pow(ZETA3,2))/3. - (3896*ZETA5)/9. + 120*EulerGamma*ZETA5)
      + CA*CA*CF*(-982.3832304526749 + (599375*EulerGamma)/729. + (50689*pow(EulerGamma,2))/162. + 
		  (4649*pow(EulerGamma,3))/81. + (121*pow(EulerGamma,4))/27. - (78607*ZETA2)/54. - 
		  (18179*EulerGamma*ZETA2)/81. - (778*pow(EulerGamma,2)*ZETA2)/9. - (88*pow(EulerGamma,3)*ZETA2)/9. + 
		  (13151*pow(ZETA2,2))/135. + (212*EulerGamma*pow(ZETA2,2))/15. + 
		  (88*pow(EulerGamma,2)*pow(ZETA2,2))/5. - (12016*pow(ZETA2,3))/315. + 
		  (115010*ZETA3)/81. - (6688*EulerGamma*ZETA3)/9. - 132*pow(EulerGamma,2)*ZETA3 + 
		  (3496*ZETA2*ZETA3)/9. + (176*EulerGamma*ZETA2*ZETA3)/3. - (248*pow(ZETA3,2))/3. - 
		  (416*ZETA5)/3. + 232*EulerGamma*ZETA5)
      + fl11dabcdabcnc*(64 + 160*ZETA2 - (32*pow(ZETA2,2))/5. + (224*ZETA3)/3. - (1280*ZETA5)/3.)
      ;
    g03 /= M_PI*M_PI*M_PI*64;
    cout << "g03 = " << g03 << endl << endl;
    //cout << nf << "   "<< g02*M_PI*M_PI*16 << "   "<< g02*M_PI*M_PI << endl;
    //cout << nf << "   "<< g03*M_PI*M_PI*M_PI*64 << "   "<< g03*M_PI*M_PI*M_PI << endl;
    //*/
    return;
  }

};






